
#from __future__ import division

import copy
from sympycore import Logic, SymbolicEquality, Calculus, heads, Symbol
from sympycore.arithmetic.numbers import div
from . import Matrix

class Polyhedron(object):
    """ A basis class that provides a H-representation for general
    polyhedra.

    Consider the following polyhedron::

      P(A_I, b_I, A_L, b_L) = { x in R^n | A_I*x <= b_I, A_L*x == b_L }

    where ``A_I`` and ``A_L`` are matrices with ``|I|`` and ``|L|`` rows,
    respectively, and ``I``, ``L`` are partitions of ``{0,..,m-1}`` such that
    ``set(I + L)==set(range(m))``, and ``*`` denotes matrix dot product.

    The class ``Polyhedron`` stores inequalities and equalities in a
    homogenized form, that is, in a m times (n+1) matrix A where the
    first column corresponds to coefficients in ``-b_I`` and ``-b_L``,
    and the rest of the columns correspond to coefficents in ``A_I``
    and ``A_L``. With respect to coefficent storage, this class
    represents the following homogeneous polyhedron::

      P(A) = { x' in R^{n+1} | A[I] * x' <= 0, A[L] * x' == 0}
    
    where ``x'`` may contain special variables, see ``labels``
    attributes for details.

    Attributes
    ----------
    A : Matrix
      A constraints matrix.

    I, L : list
      Lists of row indices of ``A``. If ``i`` in ``I`` then the
      ``i``-th row corresponds to inequality relation and if ``i`` in
      ``L`` then to equality relation. See above for definitions.

    labels : list
      List of column labels of matrix A.  Some labels have special
      meaning, as defined below:

        ``"1"`` --- the name corresponds to homogenization variable
          and the corresponding column in A corresponds to coefficents
          in ``-b_I`` and ``-b_L``. To relate P(A) to P(A_I, b_I, A_L,
          b_L), ``"1"`` value should be taken to 1.

        ``"SLACK#"`` --- the name corresponds to slack variable. The
          slack variable is used in a LP problem where it is assumed
          to be non-negative. Here ``"#"`` denotes integers used to
          index slack variables corresponding to inequality rows.

        ``"MAX"`` or ``"MIN"`` --- the labels correspond to LP
          objective function values when used in an equality relation.
          Minimization LP problem is automatically converted to
          maximization LP problem.

    objective_row, objective_column : {None, int}
      Row and column indices of the objective coefficients and ``"MAX"``
      column name, respectively.

    rhs_column : {None, int}
      Column index corresponding to RHS coefficients.
    """

    def __init__(self, *exprs):
        self.A = Matrix(0,0)
        self.I = []
        self.L = []
        self.labels = []
        self.objective_row = None
        self.objective_column = None
        self.rhs_column = None
        for expr in exprs:
            self.add(expr)

    def copy(self):
        cpy = Polyhedron()
        cpy.A = self.A.copy()
        for n in ['I', 'L', 'labels']:
            getattr(cpy, n).extend(getattr(self, n))
        for n in ['objective_row', 'objective_column']:
            setattr(cpy, n, getattr(self, n))
        return cpy
    
    @property
    def A_L(self):
        """ Return a matrix of equalities.
        """
        return self.A[tuple(self.L),:]

    @property
    def A_I(self):
        """ Return a matrix of inequalities.
        """
        return self.A[tuple(self.I),:]    

    def add(self, expr):
        """ Add extra constraint to polyhedron.

        Patameters
        ----------
        expr : {str, Logic}

          Symbolic expression of a constraint. The expression must be
          an equality or inequality relation of arbitrary linear
          combination of variables.
        """
        if isinstance(expr, str):
            if not expr or expr.strip ().startswith ('#'):
                return
            if '=' in expr or '<' in expr or '>' in expr:
                expr = Logic(expr)
            else:
                expr = Logic(expr+'==0')
        elif isinstance (expr, Calculus):
            expr = Logic (heads.EQ, (expr, Calculus(0)))
        if not isinstance(expr, Logic):
            raise NotImplementedError (`expr`) # expected Logic instance

        op, (lhs, rhs) = expr.pair
        i = self.A.rows

        if op==heads.EQ:
            expr = lhs - rhs
            self.L.append(i)
        elif op in [heads.LT, heads.LE]:
            op = heads.LE
            expr = lhs - rhs
            self.I.append(i)
        elif op in [heads.GT, heads.GE]:
            op = heads.LE
            expr = rhs - lhs
            self.I.append(i)
        else:
            raise NotImplementedError(`op`) # expected ==, <, <=, >, >=

        expr = expr.to(heads.TERM_COEFF_DICT)

        negate_row = False
        if expr.head is heads.NUMBER:
            try:
                j = self.labels.index('1')
            except ValueError:
                j = len(self.labels)
                self.labels.append(Symbol('1'))
            if self.rhs_column is None:
                self.rhs_column = j
            else:
                assert self.rhs_column == j,`self.rhs_column,j`
            self.A[i,j] = expr.data
        elif expr.head is heads.TERM_COEFF:
            term, coeff = expr.data
            if term=='MIN':
                negate_row = True
                coeff = -coeff
                term = Symbol('MAX')
            try:
                j = self.labels.index(term)
            except ValueError:
                j = len(self.labels)
                self.labels.append(term)
            if term=='MAX':
                assert self.objective_row is None, `self.objective_row, term`
                self.objective_row = i
                self.objective_column = j
            self.A[i,j] = coeff
        elif expr.head is heads.TERM_COEFF_DICT:
            for term, coeff in expr.data.iteritems():
                if term.head == heads.NUMBER:
                    assert term.data==1,`term.pair` # expected Calculus('1')
                    try:
                        j = self.labels.index('1')
                    except ValueError:
                        j = len(self.labels)
                        self.labels.append(Symbol('1'))
                    if self.rhs_column is None:
                        self.rhs_column = j
                    else:
                        assert self.rhs_column == j,`self.rhs_column,j`
                    self.A[i,j] = coeff
                else:
                    if term=='MIN':
                        negate_row = True
                        coeff = -coeff
                        term = Symbol('MAX')
                    try:
                        j = self.labels.index(term)
                    except ValueError:
                        j = len(self.labels)
                        self.labels.append(term)
                    if term=='MAX':
                        assert self.objective_row is None, `self.objective_row, term`
                        self.objective_row = i
                        self.objective_column = j
                    self.A[i,j] = coeff
        else:
            try:
                j = self.labels.index(expr)
            except ValueError:
                j = len(self.labels)
                self.labels.append(expr)
            self.A[i,j] = 1            

        self.A = self.A.resize(i+1, len (self.labels))

    def get_LP(self):
        """Construct LP dictionary matrix from polyhedra constraints.

        Parameters
        ----------
        None

        Returns
        -------
        variables : list
          List of variables.
        D : Matrix
          LP dictionary matrix.

        See also
        --------
        MatrixDict.LP_solve
        """
        assert len (self.L)==1,`self.L, self.objective_row`
        colindices = []

        variables = []
        for name, i in sorted (zip(map (str, self.labels), range (len (self.labels)))):
            if name=='1':
                pass
            elif name=='MAX':
                pass
            else:
                colindices.append(i)
                variables.append(name)
        colindices = tuple(colindices)
        rowindices = tuple(self.I)                
        D = Matrix(self.A.rows, len (colindices)+1)
        D[1:,1:] = -self.A[rowindices,colindices]
        D[0,1:] = self.A[self.objective_row, colindices] / (-self.A[self.objective_row, self.objective_column])
        if self.rhs_column is not None:
            D[:,0] = -self.A[:,self.rhs_column]

        return variables, D

    def show(self):
        print 'A:',', '.join(map(str, self.labels))
        print self.A
        print 'L:',self.L
        print 'I:',self.I




__all__ = ['POS']

from .base import ArithmeticHead

def init_module(m):
    from .base import heads
    for n,h in heads.iterNameValue(): setattr(m, n, h)

class PosHead(ArithmeticHead):

    """
    PosHead represents positive unary operation where operand (data)
    is an expression.
    """

    op_mth = '__pos__'
        
    def __repr__(self): return 'POS'

    def data_to_str_and_precedence(self, cls, expr):
        return expr.head.data_to_str_and_precedence(cls, expr.data)

    def to_lowlevel(self, cls, data, pair):
        return data

POS = PosHead()



import pandas
import numpy
import matplotlib
from matplotlib import pyplot as plt
from pymaclab import newVAR
import os


modconfig = {}
# This model's name or title, used for pickling results
modname = 'BLOOM_uncertainty_model'
modconfig['modname'] = modname
# Get the data from the directory
datafile = 'October_Updated_finaldata.csv'
modconfig['datafile'] = datafile
datapath = '../data/'
modconfig['datapath'] = datapath
data = pandas.read_csv(os.path.join(datapath,datafile),delimiter=';')
# Define frequency of data
freq = 'M'
modconfig['freq'] = freq
# Should the data be standardized?
trans_data = False
modconfig['trans_data'] = trans_data
# Define the number of lags in the VAR
nlags = 6
modconfig['nlags'] = nlags
# How many periods for IRFs ?
irfp = 48
modconfig['irfp'] = irfp
# Calculate variance decomposition?
do_vdc = True
modconfig['do_vdc'] = do_vdc
# How many periods forecast for variance decomposition ?
vdcp = 20
modconfig['vdcp'] = vdcp
# Use the svname matrix for alternative shocks?
use_svnames = False
modconfig['use_svnames'] = use_svnames
# Transform log variables in IRFs into percentages ?
translog = True
modconfig['translog'] = translog
# Transform the percentage responses back into absolutes from the mean ?
transperc = False
modconfig['transperc'] = transperc
# Add impact matrix to Cholesky responses ?
cholimp = False
modconfig['cholimp'] = cholimp
# How many matrices for MA Representation
maxphi = irfp+2
modconfig['maxphi'] = maxphi
# Boostrap IRF confidence intervals ?
bootstrap = True
modconfig['bootstrap'] = bootstrap
####################### This is experimental stuff, need to clarify the difference between the two #####################
killian_bsinbs = True
modconfig['killian_bsinbs'] = killian_bsinbs
########################################################################################################################
########################################## Graphing options for IRFs and other graphs ##################################
graph_options = {}
graph_options['labels'] = {}
graph_options['labels']['title'] = True
graph_options['labels']['xlabel'] = False
graph_options['labels']['ylabel'] = True
graph_options['labels']['text_fs'] = 10
graph_options['labels']['ylabel_fs'] = 10
graph_options['labels']['xlabel_fs'] = 10
graph_options['labels']['title_fs'] = 12
graph_options['labels']['label_fs'] = 8
graph_options['labels']['xticks_fs'] = 8
graph_options['labels']['yticks_fs'] = 8
graph_options['labels']['legend_fs'] = 8
graph_options['labels']['use_tex'] = False
graph_options['lines'] = {}
graph_options['lines']['width'] = 2.0
graph_options['grid'] = True
graph_options['irfs'] = {}
graph_options['irfs']['outer_colour'] = '#a4a4a4'
graph_options['irfs']['inner_colour'] = '#c6c6c6'
graph_options['irfs']['line_colour'] = 'black'
graph_options['irfs']['biased_line'] = True # This allows you to also show the biased IRF when Killian's b-i-b option is on
graph_options['irfs']['biased_line_type'] = 'k--'
graph_options['save'] = {}
graph_options['save']['format'] = {}
graph_options['save']['format']['eps'] = True
graph_options['save']['format']['pdf'] = True
modconfig['graph_options'] = graph_options
#########################################################################################################################
# Draw a horizontal zero line into each IRF Plot
hzline = True
modconfig['hzline'] = hzline
# How thick should the hline be?
hzline_width = 1.0
modconfig['hzline_width'] = hzline_width
# How many draws for bootstrap ?
bdraw = 1000
modconfig['bdraw'] = bdraw
# Use multicore processors to speed up bootstrap?
multicpu = True
modconfig['multicpu'] = multicpu
# Specify number of cpu cores or set to 'auto'
ncpus = 'auto'
modconfig['ncpus'] = ncpus
# Level of significance for bootstrap confidence intervals
signif = [0.66,0.95] # implies 1-alpha = 0.66, use 'mix' to graph a variety of CI, can also use 2-element list
modconfig['signif'] = signif
# What kind of intercept/time trend?
const_term = 'ct' # Can be 'None', 'cc','tt', 'ct' or 'ctt'
modconfig['const_term'] = const_term
# Pickle the model run results and settings
pickle_switch = False
# Define the names of the vars used
vnames = []
vnames.append('uncert')
vnames.append('spindex')
vnames.append('ffr')
vnames.append('employment')
vnames.append('ip')
modconfig['vnames'] = vnames
# Define the names for the vnames used in plotting
pnames = {}
pnames['uncert'] = 'Uncertainty Index'
pnames['spindex'] = 'S&P500 Index'
pnames['ffr'] = 'Federal Funds Rate'
pnames['employment'] = 'Employment'
pnames['ip'] = 'Industrial Production'
modconfig['pnames'] = pnames

# Create the data modification dictionary
tcode_dic = {}
clist = [1,4,1,4,4]
for i1,vname in enumerate(vnames):
    tcode_dic[vname] = clist[i1]
modconfig['tcode'] = tcode_dic

# Define a matrix for scaling the orthogonalized shock
svnames = {}
for name in vnames:
    svnames[name] = 1.0
svnames['uncert'] = 112.0
modconfig['svnames'] = svnames

# Set the transperc_dic
transperc_dic = {}
for name in vnames:
    transperc_dic[name] = False
transperc_dic['employment'] = False
modconfig['transperc_dic'] = transperc_dic

# Get the individual series of interest, also take natural logs where wanted
loglist = ['spindex','employment','ip']
modconfig['loglist'] = loglist
# In the list below we no longer take the log as this is done in the transform method
tlist = []
for name in vnames:
    if name in loglist:
        tlist.append(data[name].values)
    else:
        tlist.append(data[name].values)
matd = numpy.array(tlist)

# Cut off last 2 observations because of missing values
matd = matd[:,:-2]

# Transpose and get the dims
matd = matd.T
cols = matd.shape[1]
rows = matd.shape[0]
modconfig['cols'] = cols
modconfig['rows'] = rows

bvar = newVAR(data=matd,vnames=vnames,pnames=pnames,svnames=svnames,boot=True,plot=True,conf=modconfig,mesg=True)
'''
matd[:,1] = numpy.log(matd[:,1])
matd[:,3] = numpy.log(matd[:,3])
matd[:,4] = numpy.log(matd[:,4])
tsvar = var_model.VAR(matd,['uncert','spindex','ffr','emp','ip'])
tsvar_res = tsvar.fit(maxlags=6,trend='ct')
'''




__all__ = ['TERM_COEFF']

from .base import heads_precedence, ArithmeticHead, Expr

from ..core import init_module
init_module.import_heads()
init_module.import_numbers()
init_module.import_lowlevel_operations()

class TermCoeff(ArithmeticHead):
    """ Expr(TERM_COEFF, (term, coeff)) represents term*coeff
    where term is symbolic expression and coeff is a number or
    symbolic expression.
    """

    def is_data_ok(self, cls, data, allow_number_term=True):
        if type(data) is tuple and len(data)==2:
            term, coeff = data
            if isinstance(term, cls):
                if isinstance(coeff, numbertypes) or type(coeff) is not cls:
                    if term.head is self or (not allow_number_term and term.head is NUMBER):
                        return 'data[0].head cannot be %s' %(term.head)
                    return
                elif isinstance(coeff, cls):
                    if coeff.head is NUMBER:
                        if not isinstance(coeff.data, numbertypes): #pragma: no cover
                            return 'data[1].data must be %s instance for NUMBER head but got %s instance' % (numbertypes, type(coeff.data)) #pragma: no cover
                else:
                    return 'data[1] must be %s instance but got %s instance' % ((cls, numbertypes), type(coeff)) #pragma: no cover
            else:
                return 'data[0] must be %s instance but got %s instance' % (cls, type(term)) #pragma: no cover
        else:
            return 'data must be 2-tuple' #pragma: no cover
        return

    def __repr__(self): return 'TERM_COEFF'

    def new(self, cls, data):
        return term_coeff_new(cls, data)

    def reevaluate(self, cls, (term, coeff)):
        return term * coeff

    def to_EXP_COEFF_DICT(self, cls, (term, coeff), expr, variables=None):
        return term.head.to_EXP_COEFF_DICT(cls, term.data, term, variables) * coeff

    def to_MUL(self, Algebra, (term, coeff), expr):
        term = term.head.to_MUL(Algebra, term.data, term)
        if term.head is MUL:
            data = [Algebra(NUMBER, coeff)] + term.data
        else:
            data = [Algebra(NUMBER, coeff), term]
        return Algebra(MUL, data)

    def to_ADD(self, Algebra, (term, coeff), expr):
        return coeff * term.head.to_ADD(Algebra, term.data, term)

    def data_to_str_and_precedence(self, cls, (term, coeff)):
        neg_p = heads_precedence.NEG
        mul_p = heads_precedence.MUL
        if term==1:
            t, t_p = NUMBER.data_to_str_and_precedence(cls, coeff)
        elif coeff==1:
            t, t_p = term.head.data_to_str_and_precedence(cls, term.data)
        elif coeff==-1:
            t, t_p = term.head.data_to_str_and_precedence(cls, term.data)
            t, t_p = ('-('+t+')' if t_p < mul_p else '-' + t), neg_p
        elif coeff==0:
            t, t_p = '0', heads_precedence.NUMBER
        else:
            t, t_p = term.head.data_to_str_and_precedence(cls, term.data)
            c, c_p = NUMBER.data_to_str_and_precedence(cls, coeff)
            if t=='1':
                return c, c_p
            cs = '('+c+')' if c_p < mul_p else c
            ts = '('+t+')' if t_p < mul_p else t
            t = cs + (ts[1:] if ts.startswith('1/') else '*' + ts)
            t_p = mul_p
        return t, t_p

    def term_coeff(self, cls, expr):
        return expr.data

    def neg(self, cls, expr):
        term, coeff = expr.data
        return term_coeff_new(cls, (term, -coeff))

    def add(self, cls, lhs, rhs):
        term, coeff = lhs.data
        head, data = rhs.pair
        if head is ADD:
            return ADD.new(cls, [lhs] + data)
        if head is TERM_COEFF_DICT:
            data = data.copy()
            dict_add_item(cls, data, term, coeff)
            return term_coeff_dict_new(cls, data)
        if head is NUMBER:
            if data==0:
                return lhs
            return cls(TERM_COEFF_DICT,{term:coeff, cls(NUMBER,1):data})
        if head is SYMBOL:
            if term==rhs:
                return term_coeff_new(cls, (term, coeff + 1))
            return cls(TERM_COEFF_DICT,{term:coeff, rhs:1})
        if head is TERM_COEFF:
            rterm, rcoeff = data
            if rterm==term:
                return term_coeff_new(cls, (term, coeff + rcoeff))
            return cls(TERM_COEFF_DICT,{term:coeff, rterm:rcoeff})
        if head is BASE_EXP_DICT:
            rcoeff = base_exp_dict_get_coefficient(cls, data)
            if rcoeff is not None:
                d = data.copy()
                del d[rcoeff]
                rterm = base_exp_dict_new(cls, d)
                if rterm==term:
                    return term_coeff_new(cls, (term, coeff + rcoeff))
                return cls(TERM_COEFF_DICT,{term:coeff, rterm:rcoeff})
            else:
                if term==rhs:
                    return term_coeff_new(cls, (term, coeff + 1))
                return cls(TERM_COEFF_DICT,{term:coeff, rhs:1})
        if head is POW or head is APPLY or head is DIFF or head is FDIFF or head is SUBSCRIPT:
            if term==rhs:
                return term_coeff_new(cls, (term, coeff + 1))
            return cls(TERM_COEFF_DICT,{term:coeff, rhs:1})
        if head is MUL:
            return cls(ADD, [lhs, rhs])
        raise NotImplementedError(`self, rhs.head`)

    inplace_add = add

    def add_number(self, cls, lhs, rhs):
        if rhs==0:
            return lhs
        term, coeff = lhs.data
        return cls(TERM_COEFF_DICT, {term: coeff, cls(NUMBER, 1): rhs})

    def sub_number(self, cls, lhs, rhs):
        if rhs==0:
            return lhs
        term, coeff = lhs.data
        return cls(TERM_COEFF_DICT, {term: coeff, cls(NUMBER, 1): -rhs})
    
    def sub(self, cls, lhs, rhs):
        return lhs + (-rhs)

    def non_commutative_mul(self, cls, lhs, rhs):
        term, coeff = lhs.data
        head, data = rhs.pair
        if head is NUMBER:
            return term_coeff_new(cls, (term, coeff * data))
        return (term * rhs) * coeff

    commutative_mul = non_commutative_mul

    def commutative_mul_number(self, cls, lhs, rhs):
        if rhs==1:
            return lhs
        if rhs==0:
            return cls(NUMBER, 0)
        term, coeff = lhs.data
        new_coeff = coeff * rhs
        if new_coeff==1:
            return term
        return cls(TERM_COEFF, (term, new_coeff))

    non_commutative_mul_number = commutative_mul_number

    inplace_commutative_mul = commutative_mul

    def commutative_div_number(self, cls, lhs, rhs):
        term, coeff = lhs.data
        r = number_div(cls, coeff, rhs)
        if rhs==0:
            return r * term
        return term_coeff_new(cls, (term, r))

    def commutative_rdiv_number(self, cls, lhs, rhs):
        term, coeff = lhs.data
        r = number_div(cls, rhs, coeff)
        if rhs==0:
            return r * (1/term)
        return term_coeff_new(cls, (1/term, r))

    def commutative_div(self, cls, lhs, rhs):
        rhead, rdata = rhs.pair
        if rhead is NUMBER:
            return self.commutative_div_number(cls, lhs, rdata)
        term, coeff = lhs.data
        if rhead is TERM_COEFF:
            rterm, rcoeff = rdata
            if term==rterm:
                return cls(NUMBER, number_div(cls, coeff, rcoeff))
            return term.head.commutative_div(cls, term, rterm) * number_div(cls, coeff, rcoeff)
        if rhead is SYMBOL or rhead is APPLY:
            if term == rhs:
                return cls(NUMBER, coeff)
            b, e = term.base_exp()
            return cls(TERM_COEFF, (cls(BASE_EXP_DICT, {b:e, rhs:-1}), coeff))
        if rhead is TERM_COEFF_DICT:
            b, e = term.base_exp()
            return cls(TERM_COEFF, (cls(BASE_EXP_DICT, {b:e, rhs:-1}), coeff))
        if term==rhs:
            return cls(NUMBER, coeff)
        return (term / rhs) * coeff
        return term.head.commutative_div(cls, term, rhs) * coeff

    def pow(self, cls, base, exp):
        term, coeff = base.data
        if isinstance(exp, Expr):
            head, data = exp.pair
            if head is NUMBER:
                exp = data
        if isinstance(exp, inttypes):
            if exp<0:
                return term ** exp / coeff ** (-exp)
            return term ** exp * coeff ** exp
        return pow_new(cls, (base, exp))

    def pow_number(self, cls, base, exp):
        term, coeff = base.data
        if isinstance(exp, inttypes):
            if exp<0:
                return term ** exp / coeff ** (-exp)
            return term ** exp * coeff ** exp
        return cls(POW, (base, exp))

    def walk(self, func, cls, data, target):
        term, coeff = data
        h, d = term.pair
        term = h.walk(func, cls, d, term)
        if isinstance(coeff, Expr):
            h, d = coeff.pair
            coeff = h.walk(func, type(coeff), d, coeff)
        s = term * coeff
        h, d = s.pair
        return func(cls, h, d, s)

    def scan(self, proc, cls, data, target):
        term, coeff = data
        term.head.scan(proc, cls, term.data, target)    
        if isinstance(coeff, Expr):
            coeff.head.scan(proc, type(coeff), coeff.data, target)
        proc(cls, self, data, target)

    def expand(self, cls, expr):
        term, coeff  = expr.data
        return term.head.expand(cls, term) * coeff

    def diff(self, cls, data, expr, symbol, order, cache={}):
        term, coeff = data
        return term.head.diff(cls, term.data, term, symbol, order, cache=cache) * coeff


    def diff_apply(self, cls, data, diff, expr):
        term, coeff = data
        return term.head.diff_apply(cls, term.data, term, expr) * coeff

    def apply(self, cls, data, func, args):
        term, coeff = data
        return term.head.apply(cls, term.data, term, args) * coeff

    def integrate_indefinite(self, cls, data, expr, x):
        term, coeff = data
        return term.head.integrate_indefinite(cls, term.data, term, x) * coeff

    def integrate_definite(self, cls, data, expr, x, a, b):
        term, coeff = data
        return term.head.integrate_definite(cls, term.data, term, x, a, b) * coeff

    def to_TERM_COEFF_DICT(self, Algebra, data, expr):
        term, coeff = data
        return term.head.to_TERM_COEFF_DICT(Algebra, term.data, term) * coeff

    def algebra_pos(self, Algebra, expr):
        return expr

    def algebra_neg(self, Algebra, expr):
        if Algebra.algebra_options.get('evaluate_addition'):
            term, coeff = expr.data
            return term_coeff_new(Algebra, (term, -coeff))
        return Algebra(NEG, expr)

    def algebra_add_number(self, Algebra, lhs, rhs, inplace):
        return self.algebra_add(Algebra, lhs, Algebra(NUMBER, rhs), inplace)

    def algebra_add(self, Algebra, lhs, rhs, inplace):
        rhead, rdata = rhs.pair
        if Algebra.algebra_options.get('is_additive_group_commutative'):
            lterm, lcoeff = lhs.data
            if Algebra.algebra_options.get('evaluate_addition'):
                if rhead is ADD or rhead is MUL or rhead is NEG:
                    rhs = rhead.to_TERM_COEFF_DICT(Algebra, rdata, rhs)
                    rhead, rdata = rhs.pair
                if rhead is NUMBER:
                    if rdata==0:
                        return lhs
                    return Algebra(TERM_COEFF_DICT, {lterm:lcoeff, Algebra(NUMBER, 1): rdata})
                if rhead is SYMBOL:
                    if lterm==rhs:
                        return term_coeff_new(Algebra, (lterm, lcoeff + 1))
                    return Algebra(TERM_COEFF_DICT, {lterm:lcoeff, rhs: 1})
                if rhead is TERM_COEFF:
                    rterm, rcoeff = rdata
                    if lterm == rterm:
                        return term_coeff_new(Algebra, (lterm, lcoeff+rcoeff))
                    return Algebra(TERM_COEFF_DICT, {lterm:lcoeff, rterm:rcoeff})
                if rhead is TERM_COEFF_DICT:
                    d = rdata.copy()
                    term_coeff_dict_add_item(Algebra, d, lterm, lcoeff)
                    return term_coeff_dict_new(Algebra, d)
            else:
                if rhead is TERM_COEFF_DICT or rhead is EXP_COEFF_DICT:
                    rhead, rdata = rhs.to(ADD).pair
                if rhead is ADD:
                    data = [lhs] + rdata
                else:
                    data = [lhs, rhs]
                return add_new(Algebra, data)
            return super(type(self), self).algebra_add(Algebra, lhs, rhs, inplace)
        else:
            if rhead is TERM_COEFF_DICT or rhead is EXP_COEFF_DICT:
                rhead, rdata = rhs.to(ADD).pair
            if rhead is ADD:
                data = [lhs] + rdata
            else:
                data = [lhs, rhs]
            if Algebra.algebra_options.get('evaluate_addition'):
                ADD.combine_add_list(Algebra, data)
            return add_new(Algebra, data)

    def algebra_mul_number(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_addition'):
            if not rhs:
                return Algebra(NUMBER, 0)
            term, coeff = lhs.data
            return term_coeff_new(Algebra, (term, coeff * rhs))
        return mul_new(Algebra, [lhs, Algebra(NUMBER, rhs)])

    def algebra_mul(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_addition'):
            term, coeff = lhs.data
            rhead, rdata = rhs.pair
            if rhead is NUMBER:
                return term_coeff_new(Algebra, (term, coeff*rdata))
            return super(type(self), self).algebra_mul(Algebra, lhs, rhs, inplace)
        return mul_new(Algebra, [lhs, rhs])

TERM_COEFF = TermCoeff()



#!/usr/bin/env python
import os
import sys
import shutil
import glob


######################################################################
# Check dependencies and install using pip
######################################################################
# Now check for numpy and install if needed
numpy_version = False
try:
    import numpy
    numpy_version = numpy.__version__
    if numpy_version and numpy_version != '1.6.2':
        print "You need Numpy version 1.6.2 for PyMacLab. Re-installing correct version now..."
        os.system("pip install numpy==1.6.2")    
except:
    print "Numpy not detected. Fetching version 1.6.2 now using pip..."
    os.system("pip install numpy==1.6.2")
    
# Now check for scipy and install if needed
scipy_version = False
try:
    import scipy
    scipy_version = scipy.__version__
    if scipy_version and scipy_version != '0.11.0':
        print "You need Scipy version 0.11.0 for PyMacLab. Re-installing correct version now..."
        os.system("pip install scipy==0.11.0") 
except:
    print "Scipy not detected. Fetching version 0.11.0 now using pip..."
    os.system("pip install scipy==0.11.0")

# Now check for matplotlib and install if needed
matplotlib_version = False
try:
    import matplotlib
    matplotlib_version = matplotlib.__version__
    if matplotlib_version and matplotlib_version != '1.1.0':
        print "You need Matplotlib version 1.1.0 for PyMacLab. Re-installing correct version now..."
        os.system("pip install matplotlib==1.1.0") 
except:
    print "Matplotlib not detected. Fetching now using pip..."
    os.system("pip install matplotlib==1.1.0")

# Now check for pandas and install if needed
pandas_version = False
try:
    import pandas
    pandas_version = pandas.__version__
    if pandas_version and pandas_version != '0.9.0':
        print "You need Pandas version 0.9.0 for PyMacLab. Re-installing correct version now..."
        os.system("pip install pandas==0.9.0") 
except:
    print "Pandas not detected. Fetching now using pip..."
    os.system("pip install pandas==0.9.0")
    
# Now check for ipython and install if needed
ipython_version = False
try:
    import IPython
    ipython_version = IPython.__version__
    if ipython_version and ipython_version != '0.13':
        print "You should use IPython version 0.13 for PyMacLab. Re-installing correct version now..."
        os.system("pip install ipython==0.13") 
except:
    print "IPython not detected but is fun to use with PyMacLab. Fetching now using pip..."
    os.system("pip install ipython==0.13")

'''
# Now check for pp and install if needed
pp_version = False
try:
    import pp
    pp_version = pp.version
    if pp_version and pp_version != '1.6.2':
        print "You should use PP version 1.6.2 for PyMacLab. Re-installing correct version now..."
        os.system("pip install pp==1.6.2")
except:
    print "PP not detected but makes PyMacLab much faster. Fetching now using pip..."
    os.system("pip install pp==1.6.2")
'''

# Now check for wheezy.template and install if needed, but cannot check version so just install any latest
try:
    import wheezy.template
except:
    print "Wheezy.template not detected but used in PyMacLab. Fetching now using pip..."
    os.system("pip install wheezy.template")
##########################################################################
# Done checking for dependencies
##########################################################################
    

from datetime import datetime

# import setuptools # have to do this to be able to setup.py develop
from numpy.distutils.core import setup
from numpy import get_include

# Remove the build directory
if 'build' in os.listdir(os.getcwd()): shutil.rmtree(os.path.join(os.getcwd(),'build'))

DESCRIPTION="The Python Macroeconomics Library"
LONG_DESCRIPTION="""
PyMacLab stands for Python Macroeconomics Library which currently primarily serves the purposes of providing
a convenience framework written in Python to solve non-linear DSGE models. At the time of writing this the library
supports solving DSGE models using 1st and 2nd order perturbation methods which are computed around the steady state.
In particular, the library provides wrapper function for [Paul Klein's](http://paulklein.ca/newsite/start/start.php)
1st-order accurate method based on the Schur Decomposition as well a more recently published method by the same author
(co-authored with Paul Gomme, see [here](http://ideas.repec.org/a/eee/dyncon/v35y2011i4p604-615.html)) which provides
2nd-order accurate solutions without using Tensor Algebra (using the Magnus and Neudecker 1999 definition of the
Hessian matrix).

The library is extremely user-friendly in the sense of employing a model text file parser similar to that present in
[Dynare](http://www.dynare.org/) which requires users to only write down the original set of non-linear first-order
conditions of optimality. In addition, users are offered a menu of options of how to provide information required for
calculating the steady state of the model. Once the model is parsed and read in, several options of solving it exist
and users are provided with further convenience methods suitable for simulating solved models and investigating dynamic
statistical properties.

It should also be mentioned that because PyMacLab is a convenience library of highly modular nature (using
a object-oriented programming approach) it is very easy to loop over one model several thousand times each time changing
the original set of primitive parameters, such as depreciation rates, impatience factors, etc. in order to compute
solutions of the same model over a large set of conceivable parameterisations. Also, whenever solution methods require
the calculation of the Jacobian or Hessian, this is always done analytically (symbolically) using the Python
symbolic computation library [SympyCore](http://code.google.com/p/sympycore/) and not numerically as in other software
packages. Sympycore is not supplanted by Sympy, but it works well at the moment so we will alter PyMacLab at a later
stage to reflect this.

PyMacLab was authored by [Eric M. Scheffel](http://www.ericscheffel.com) who is currently working as [Assistant Professor
in Economics at Nottingham University China](http://www.nottingham.edu.cn/en/business/people/staffprofile/eric-scheffel.aspx)
and is distributed under the GNU General Public License v3.0.
"""

DISTNAME = 'pymaclab'
LICENSE ="""
Copyright 2007-2012 Eric M. Scheffel

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
AUTHOR = "Eric M. Scheffel"
MAINTAINER = "Eric M. Scheffel"
MAINTAINER_EMAIL = "emscheffel@gmail.com"
URL = 'http://github.com/escheffel/pymaclab/'
DOWNLOAD_URL="http://github.com/escheffel/pymaclab/tarball/v0.90.1"
CLASSIFIERS=["Scientific", "Macroeconomics", "General Equilibrium", "DSGE", "Time Series"]

MAJOR = 0
MINOR = 90
MICRO = 1
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)
FULLVERSION = VERSION
if not ISRELEASED:
    FULLVERSION += '.dev'

def write_version_py(filename='./version.py'):
    cnt = """\
from datetime import datetime

version = '%s'
"""
    a = open(filename, 'w')
    try:
        a.write(cnt % FULLVERSION)
    finally:
        a.close()

def configuration(parent_package='', top_path=None, package_name=DISTNAME):
#    write_version_py()

    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path,
                           version=FULLVERSION)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage('pymaclab')
    return config

if __name__ == '__main__':
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          download_url=DOWNLOAD_URL,
          long_description=LONG_DESCRIPTION,
          classifiers=CLASSIFIERS,
          platforms='any',
          configuration=configuration)
    '''
    # Find the standard site-packages absolute path
    for elem in sys.path:
        if 'site-packages' in elem: pathos = os.path.join(elem.split('site-packages')[0],'site-packages','pymaclab')
    '''
    # Install pp and sympycore, but first remove old stuff
    # Delete old sympycore stuff
    try:
        # We don't want to import the package in the current working directory!
        sys.path.pop(0)
        import sympycore as sc
        patho = sc.__path__
        shutil.rmtree(patho[0])
        globo = glob.glob(patho[0]+'*')
        for elem in globo:
            lasto = elem.split(r'/')[-1]
            if lasto in os.listdir(elem.split(lasto)[0]):
                shutil.rmtree(elem)
    except:
        pass
    # Delete old pp stuff
    try:
        # We don't want to import the package in the current working directory!
        sys.path.pop(0)
        import pp as ppp
        filo = ppp.__file__
        nameos = ['pptransport.py','pptransport.pyc',
                  'ppauto.py','ppauto.pyc',
                  'ppcommon.py','ppcommon.pyc',
                  'pp.py','pp.pyc',
                  'ppworker.py','ppworker.pyc',
                  'pp-*.egg-info']
        diro = filo.split('pp.')[0]
        for elem in nameos:
            if '*' not in nameos:
                if elem in os.listdir(diro):
                    shutil.rmtree(os.path.join(diro,elem))
            else:
                globo = glob.glob(os.path.join(diro,elem))
                for elem2 in globo:
                    lasto = elem2.split(r'/')[-1]
                    if lasto in os.listdir(elem2.split(lasto)[0]):
                        shutil.rmtree(elem2)
        shutil.rmtree(patho[0])
    except:
        pass
    
    # Now insert the current directory back in
    sys.path[0] = ''    
    os.system("python setup_pp.py install")
    os.system("python setup_sympycore.py install")
    



__all__ = ['HEAD', 'SYMBOL', 'NUMBER', 'SPECIAL', 'CALLABLE',
           'ADD', 'SUB', 'MUL', 'MOD', 'DIV', 'FLOORDIV', 'POW', 'POS', 'NEG',
           'POW', 'TERMS', 'FACTORS',
           'EQ', 'NE', 'LT', 'GT', 'LE', 'GE',
           'INVERT', 'BOR', 'BXOR', 'BAND', 'LSHIFT', 'RSHIFT',
           'APPLY', 'SUBSCRIPT', 'SLICE', 'LAMBDA', 'ATTR', 'KWARG',
           'NOT', 'OR', 'AND', 'IS', 'ISNOT', 'IN', 'NOTIN',
           'TUPLE', 'LIST', 'DICT',
           'SPARSE_POLY', 'DENSE_POLY',
           'EXP_COEFF_DICT',
           'TERM_COEFF_DICT',
           'TERM_COEFF',
           'BASE_EXP_DICT',
           'INTEGRAL_DEFINITE', 'INTEGRAL_INDEFINITE'
           ]

from .base import HEAD

from .symbol import SYMBOL
from .special import SPECIAL
from .number import NUMBER
from .callable import CALLABLE

from .pos import POS
from .neg import NEG
from .add import ADD
from .sub import SUB
from .pow import POW
from .mul import MUL
from .div import DIV
from .term_coeff_dict import TERM_COEFF_DICT
from .term_coeff_dict import TERM_COEFF_DICT as TERMS
from .base_exp_dict import BASE_EXP_DICT
from .base_exp_dict import BASE_EXP_DICT as FACTORS
from .term_coeff import TERM_COEFF

from .apply import APPLY

from .differential import DIFF, FDIFF

from .integral_indefinite import INTEGRAL_INDEFINITE
from .integral_definite import INTEGRAL_DEFINITE

from .arithmetic import MOD, FLOORDIV
from .relational import EQ, NE, LT, GT, LE, GE
from .binary import INVERT, BOR, BXOR, BAND, LSHIFT, RSHIFT
from .functional import SUBSCRIPT, SLICE, LAMBDA, ATTR, KWARG
from .logic import NOT, OR, AND, IS, ISNOT, IN, NOTIN
from .containers import TUPLE, LIST, DICT
from .polynomial import SPARSE_POLY, DENSE_POLY

from .exp_coeff_dict import EXP_COEFF_DICT
from .integer_list import INTEGER_LIST




from sympycore import Calculus as Algebra

Symbol = Algebra.Symbol
Number = Algebra.Number
Add = Algebra.Add
Mul = Algebra.Mul
Pow = Algebra.Pow
Terms = Algebra.Terms
Factors = Algebra.Factors
Apply = Algebra.Apply

x,y,z = map(Symbol, 'xyz')

def foo(x):
    """ Python function can be used to construct applied functions.
    """
    if x==0:
        return x
    return Apply(foo, x)


class bar(object):
    """ Python new-style class can be used to construct applied
    functions.
    """
    def __new__(cls, x):
        if x==0:
            return x
        r = Apply(cls, x)
        return r

class Fun:
    """ A callable class instance can be used to construct applied
    functions.
    """
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name
    
    def __call__(self, x):
        if x==0:
            return x
        return Apply(self, x)

fun = Fun('fun')

def mysin(x):
    if x==0:
        return x
    return Apply(mysin, x)


def mycos(x):
    if x==0:
        return Algebra.one
    return Apply(mycos, x)


mysin.derivative = lambda arg: mycos(arg)
mycos.derivative = lambda arg: -mysin(arg)

mysin.fdiff = lambda fcls, arg_index, order: (lambda x: mycos(x))
mycos.fdiff = lambda fcls, arg_index, order: (lambda x: -mysin(x))

def foo2(x, y):
    if x==y:
        return Algebra.one
    return Apply(foo2, x,y)

def foo2_1(x, y):
    return Apply(foo2_1, x,y)

def foo2_2(x, y):
    return Apply(foo2_2, x,y)


foo2.derivative = foo2_1,foo2_2

def foo2_fdiff(fcls, arg_index, order):
    assert order==1 and arg_index in [0,1], `arg_index, order`
    if arg_index==0: return foo2_1
    return foo2_2

foo2.fdiff = foo2_fdiff

def bar2(x, y):
    return Apply(bar2, x,y)


def test_foo():
    assert str(foo(x))=='foo(x)'
    assert str(foo(0))=='0'
    assert str(foo(x+y)) in ['foo(x + y)', 'foo(y + x)']

    assert foo(x)==foo(x)
    assert foo(x).subs(x,z)==foo(z)
    assert foo(x).subs(x,0)==0
    assert (y+foo(x)).subs(x,0)==y

    assert str(foo(foo(x)))=='foo(foo(x))'
    assert str(foo(x)+foo(y))=='foo(x) + foo(y)'
    assert str(foo(x)+x*foo(y)) in ['foo(x) + x*foo(y)', 'x*foo(y) + foo(x)'],  str(foo(x)+x*foo(y))

    assert str((foo(x)+foo(y))**2)=='(foo(x) + foo(y))**2'
    assert str(((foo(x)+foo(y))**2).expand()) in ['2*foo(x)*foo(y) + foo(x)**2 + foo(y)**2',
                                                  'foo(x)**2 + foo(y)**2 + 2*foo(x)*foo(y)',
                                                  'foo(y)**2 + foo(x)**2 + 2*foo(x)*foo(y)',
                                                  'foo(x)**2 + 2*foo(x)*foo(y) + foo(y)**2',
                                                  'foo(y)**2 + 2*foo(x)*foo(y) + foo(x)**2'], str(((foo(x)+foo(y))**2).expand())


def test_bar():
    assert str(bar(x))=='bar(x)'
    assert str(bar(0))=='0'
    assert str(bar(x+y)) in ['bar(x + y)', 'bar(y + x)']

    assert bar(x)==bar(x)
    assert bar(x).subs(x,z)==bar(z), `bar(x).subs(x,z), bar(z)`
    assert bar(x).subs(x,0)==0
    assert (y+bar(x)).subs(x,0)==y

    assert str(bar(bar(x)))=='bar(bar(x))'
    assert str(bar(x)+bar(y))=='bar(x) + bar(y)',  str(bar(x)+bar(y))
    assert str(bar(x)+x*bar(y)) in ['bar(x) + x*bar(y)','x*bar(y) + bar(x)'], str(bar(x)+x*bar(y))

    assert str(((bar(x)+bar(y))**2).expand()) in ['2*bar(x)*bar(y) + bar(x)**2 + bar(y)**2',
                                                  'bar(x)**2 + bar(y)**2 + 2*bar(x)*bar(y)',
                                                  'bar(y)**2 + bar(x)**2 + 2*bar(x)*bar(y)',
                                                  'bar(x)**2 + 2*bar(x)*bar(y) + bar(y)**2',
                                                  'bar(y)**2 + 2*bar(x)*bar(y) + bar(x)**2'], str(((bar(x)+bar(y))**2).expand())

def test_fun():
    assert str(fun(x))=='fun(x)'
    assert str(fun(0))=='0'
    assert str(fun(x+y)) in ['fun(x + y)', 'fun(y + x)'], str(fun(x+y))

    assert fun(x)==fun(x)
    assert fun(x).subs(x,z)==fun(z)
    assert fun(x).subs(x,0)==0
    assert (y+fun(x)).subs(x,0)==y

    assert str(fun(fun(x)))=='fun(fun(x))'
    assert str(fun(x)+fun(y))=='fun(x) + fun(y)'
    assert str(fun(x)+x*fun(y)) in ['fun(x) + x*fun(y)', 'x*fun(y) + fun(x)'], str(fun(x)+x*fun(y))

    assert str(((fun(x)+fun(y))**2).expand()) in ['2*fun(x)*fun(y) + fun(x)**2 + fun(y)**2',
                                                  'fun(x)**2 + fun(y)**2 + 2*fun(x)*fun(y)',
                                                  'fun(y)**2 + fun(x)**2 + 2*fun(x)*fun(y)',
                                                  'fun(x)**2 + 2*fun(x)*fun(y) + fun(y)**2',
                                                  'fun(y)**2 + 2*fun(x)*fun(y) + fun(x)**2'],\
                                                  str(((fun(x)+fun(y))**2).expand())

def test_diff():
    assert str(mysin(x))=='mysin(x)'
    assert str(mysin(x).diff(x))=='mycos(x)'
    assert str(mycos(x).diff(x))=='-mysin(x)'
    assert str(mysin(mysin(x)).diff(x)) in ['mycos(x)*mycos(mysin(x))','mycos(mysin(x))*mycos(x)']

def test_foo2():
    assert str(foo2(x,x))=='1'
    assert str(foo2(x,y))=='foo2(x, y)'
    assert str(foo2(x,y).subs(x,z))=='foo2(z, y)'

    assert str(foo2(x,y).diff(x))=='foo2_1(x, y)'
    assert str(foo2(x,y).diff(y))=='foo2_2(x, y)'
    assert str(foo2(foo(x),bar(x)).diff(x)) in ['bar_1(x)*foo2_2(foo(x), bar(x)) + foo_1(x)*foo2_1(foo(x), bar(x))',
                                                'bar_1(x)*foo2_2(foo(x), bar(x)) + foo2_1(foo(x), bar(x))*foo_1(x)',
                                                'foo_1(x)*foo2_1(foo(x), bar(x)) + bar_1(x)*foo2_2(foo(x), bar(x))',
                                                'FD[0](foo)(x)*foo2_1(foo(x), bar(x)) + FD[0](bar)(x)*foo2_2(foo(x), bar(x))'],\
                                                str(foo2(foo(x),bar(x)).diff(x))

def test_bar2():
    assert str(bar2(x,y))=='bar2(x, y)'
    assert str(bar2(x,y).subs(x,z))=='bar2(z, y)'
    assert str(bar2(x,y).diff(x)) in ['bar2_1(x, y)','FD[0](bar2)(x, y)'],str(bar2(x,y).diff(x))
    assert str(bar2(x,y).diff(y)) in ['bar2_2(x, y)','FD[1](bar2)(x, y)'], str(bar2(x,y).diff(y))
    assert str(bar2(x,x).diff(x)) in ['bar2_1(x, x) + bar2_2(x, x)',
                                      'bar2_2(x, x) + bar2_1(x, x)',
                                      'FD[1](bar2)(x, x) + FD[0](bar2)(x, x)'],\
                                      str(bar2(x,x).diff(x))



# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2012, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, Execution Server

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""

import threading

copyright = "Copyright (c) 2005-2012 Vitalii Vanovschi. All rights reserved"
version = "1.6.2"

def start_thread(name,  target,  args=(),  kwargs={},  daemon=True):
    """Starts a thread"""
    thread = threading.Thread(name=name,  target=target, args=args,  kwargs=kwargs)
    thread.daemon = daemon
    thread.start()
    return thread


def get_class_hierarchy(clazz):
    classes = []
    if clazz is type(object()):
        return classes
    for base_class in clazz.__bases__:
        classes.extend(get_class_hierarchy(base_class))
    classes.append(clazz)
    return classes


def is_not_imported(arg, modules):
    args_module = str(arg.__module__)
    for module in modules:
        if args_module == module or args_module.startswith(module + "."):
            return False
    return True

# Parallel Python Software: http://www.parallelpython.com




import time
from sympycore import Polyhedron, Matrix, mpq


def test_LP_solve_Chateau_ETH_Production():
    constraints = '''\
3*x1+4*x2+2*x3==MAX
2*x1<=4
x1+2*x3<=8
3*x2+x3<=6
'''
    p = Polyhedron(*constraints.split('\n'))
    #p.show()
    names, D = p.get_LP()
    Dwork = D[:]
    xopt, vopt = Dwork.LP_solve(overwrite=True)
    assert (D[0,1:] * xopt)[0,0]==vopt==16,`D[0,1:] * xopt,vopt`

def test_LP_solve_cyclic():
    constraints = '''
x1-2*x2+x3==MAX
2*x1-x2+x3<=0
3*x1+x2+x3<=0
-5*x1+3*x2-2*x3<=0
'''
    p = Polyhedron(*constraints.split('\n'))
    #p.show()
    names, D = p.get_LP()
    Dwork = D[:]
    xopt, vopt = Dwork.LP_solve(overwrite=True)
    assert (D[0,1:] * xopt)[0,0]==vopt==0,`D[0,1:] * xopt,vopt`

def test_LP_solve_Karmarkar_example ():
    p = Polyhedron('x1+x2==MAX')
    for i in range (11):
        p1 = mpq((i,10))*1
        p.add('2*%s*x1+x2<=%s+1' % (p1,p1**2))
    names, D = p.get_LP()
    Dwork = D[:]
    xopt, vopt = Dwork.LP_solve(overwrite=True)
    assert (D[0,1:] * xopt)[0,0]==vopt==mpq ((5,4)),`D[0,1:] * xopt,vopt`

def test_Av98a():
    constraints = '''\
#x1+x2+x3==1
1+x1>=0
1+x2>=0
1-x1>=0
1-x2>=0
1-x1+x3>=0
1-x2+x3>=0
1+x1+x3>=0
1+x2+x3>=0'''
    p = Polyhedron(*constraints.split('\n'))
    #p.show()

    return

    vertices = [(1,1,0),(-1,1,0),(1,-1,0),(-1,-1,0),(0,0,-1)]
    rays = [(0,0,1)]

    print 'A_L'
    print p.A_L

    xd, dep, indep = p.A_L.solve_null(labels=p.names)
    Aker = Matrix([xd[l].tolist()[0] for l in p.names])
    print 'Aker:'
    print Aker

    print 'A * Aker:'
    print p.A * Aker

    xd, dep, indep = p.A_I[(0,1),:].solve_null(labels=p.names)
    Aker = Matrix([xd[l].tolist()[0] for l in p.names])
    print 'Aker:'
    print Aker
    print 'A * Aker:'
    print p.A * Aker

    print p.A[(0,1,2,4),:].I*2
    print p.A[(0,1,2,4),:]*p.A[(0,1,2,4),:].I



from sympycore.arithmetic.numbers import normalized_fraction as mpq
from sympycore.arithmetic.numbers import mpf
from sympycore.arithmetic.number_theory import *
from sympycore.arithmetic.number_theory import binomial_coefficients, binomial_coefficients_list

def test_factorial():
    assert factorial(0) == 1
    assert factorial(1) == 1
    assert factorial(2) == 2
    assert factorial(10) == 3628800
    # Verify that memoization didn't cause trouble
    assert factorial(10) == 3628800
    assert factorial(130) % 1009 == 926

def test_gcd():
    assert gcd() == 0
    assert gcd(2) == 2
    assert gcd(2,3) == 1
    assert gcd(2*3*5, 2*3*7) == 6
    assert gcd(2*3*5, 2*5, 3*5) == 5
    assert gcd(mpq(2,3),mpq(5,8)) == mpq(1,24)

def test_lcm():
    assert lcm(0) == 0
    assert lcm(1) == 1
    assert lcm(12,18) == 36
    assert lcm(7,11,7) == 77
    assert lcm(12,16,15,9) == 720
    assert lcm(mpq(1,2),mpq(5,6)) == mpq(5,2)

def test_integer_digits():
    assert integer_digits(0) == [0]
    assert integer_digits(1234, 10) == [1,2,3,4]
    assert integer_digits(37, 2) == [1,0,0,1,0,1]

def test_real_digits():
    assert real_digits(0) == ([0], 1)
    assert real_digits(28) == ([2,8], 2)
    assert real_digits(mpq(3,2)) == ([1,5], 1)
    assert real_digits(mpq(1,1000)) == ([1], -2)
    assert real_digits(mpq(355,113)) == ([3,1,4,1,5,9,2,9,2,0], 1)
    assert real_digits(mpq(1,10),2) == ([1,1,0,0,1,1,0,0,1,1], -3)
    # 1234.1234...
    p = mpq(1234123456789,1000000000)
    assert real_digits(p, 10, 8) == ([1,2,3,4,1,2,3,4],4)
    assert real_digits(p, 10, 12) == ([1,2,3,4,1,2,3,4,5,6,7,8], 4)

def test_binomial_coefficients_list():
    assert binomial_coefficients_list(0) == [1]
    assert binomial_coefficients_list(1) == [1,1]
    assert binomial_coefficients_list(2) == [1,2,1]
    assert binomial_coefficients_list(3) == [1,3,3,1]
    assert binomial_coefficients_list(4) == [1,4,6,4,1]
    assert binomial_coefficients_list(5) == [1,5,10,10,5,1]
    assert binomial_coefficients_list(6) == [1,6,15,20,15,6,1]

def test_binomial_coefficients():
    for n in range(15):
        c = binomial_coefficients(n)
        l = [c[k] for k in sorted(c)]
        assert l==binomial_coefficients_list(n)

def test_f2q():
    assert f2q(mpf(20)/521)==mpq(20, 521)
    assert f2q(mpf(1)/3)==mpq(1, 3), `f2q(mpf(1)/3)`
    assert f2q(mpf(1232)/5224)==mpq(154, 653), `f2q(mpf(1232)/5224)`



"""
This module implements computation of elementary transcendental
functions (powers, logarithms, trigonometric and hyperbolic
functions, inverse trigonometric and hyperbolic) for real
floating-point numbers.

For complex and interval implementations of the same functions,
see libmpc and libmpi.

"""

import math
from bisect import bisect

from .backend import xrange
from .backend import MPZ, MPZ_ZERO, MPZ_ONE, MPZ_TWO, MPZ_FIVE, BACKEND

from .libmpf import (
    round_floor, round_ceiling, round_down, round_up,
    round_nearest, round_fast,
    ComplexResult,
    bitcount, bctable, lshift, rshift, giant_steps, sqrt_fixed,
    from_int, to_int, from_man_exp, to_fixed, to_float, from_float,
    from_rational, normalize,
    fzero, fone, fnone, fhalf, finf, fninf, fnan,
    mpf_cmp, mpf_sign, mpf_abs,
    mpf_pos, mpf_neg, mpf_add, mpf_sub, mpf_mul, mpf_div, mpf_shift,
    mpf_rdiv_int, mpf_pow_int, mpf_sqrt,
    reciprocal_rnd, negative_rnd, mpf_perturb,
    isqrt_fast
)

from .libintmath import ifib


#-------------------------------------------------------------------------------
# Tuning parameters
#-------------------------------------------------------------------------------

# Cutoff for computing exp from cosh+sinh. This reduces the
# number of terms by half, but also requires a square root which
# is expensive with the pure-Python square root code.
if BACKEND == 'python':
    EXP_COSH_CUTOFF = 600
else:
    EXP_COSH_CUTOFF = 400
# Cutoff for using more than 2 series
EXP_SERIES_U_CUTOFF = 1500

# Also basically determined by sqrt
if BACKEND == 'python':
    COS_SIN_CACHE_PREC = 400
else:
    COS_SIN_CACHE_PREC = 200
COS_SIN_CACHE_STEP = 8
cos_sin_cache = {}

# Number of integer logarithms to cache (for zeta sums)
MAX_LOG_INT_CACHE = 2000
log_int_cache = {}

LOG_TAYLOR_PREC = 2500  # Use Taylor series with caching up to this prec
LOG_TAYLOR_SHIFT = 9    # Cache log values in steps of size 2^-N
log_taylor_cache = {}
# prec/size ratio of x for fastest convergence in AGM formula
LOG_AGM_MAG_PREC_RATIO = 20

ATAN_TAYLOR_PREC = 3000  # Same as for log
ATAN_TAYLOR_SHIFT = 7   # steps of size 2^-N
atan_taylor_cache = {}


# ~= next power of two + 20
cache_prec_steps = [22,22]
for k in xrange(1, bitcount(LOG_TAYLOR_PREC)+1):
    cache_prec_steps += [min(2**k,LOG_TAYLOR_PREC)+20] * 2**(k-1)


#----------------------------------------------------------------------------#
#                                                                            #
#                   Elementary mathematical constants                        #
#                                                                            #
#----------------------------------------------------------------------------#

def constant_memo(f):
    """
    Decorator for caching computed values of mathematical
    constants. This decorator should be applied to a
    function taking a single argument prec as input and
    returning a fixed-point value with the given precision.
    """
    f.memo_prec = -1
    f.memo_val = None
    def g(prec, **kwargs):
        memo_prec = f.memo_prec
        if prec <= memo_prec:
            return f.memo_val >> (memo_prec-prec)
        newprec = int(prec*1.05+10)
        f.memo_val = f(newprec, **kwargs)
        f.memo_prec = newprec
        return f.memo_val >> (newprec-prec)
    g.__name__ = f.__name__
    g.__doc__ = f.__doc__
    return g

def def_mpf_constant(fixed):
    """
    Create a function that computes the mpf value for a mathematical
    constant, given a function that computes the fixed-point value.

    Assumptions: the constant is positive and has magnitude ~= 1;
    the fixed-point function rounds to floor.
    """
    def f(prec, rnd=round_fast):
        wp = prec + 20
        v = fixed(wp)
        if rnd in (round_up, round_ceiling):
            v += 1
        return normalize(0, v, -wp, bitcount(v), prec, rnd)
    f.__doc__ = fixed.__doc__
    return f

def bsp_acot(q, a, b, hyperbolic):
    if b - a == 1:
        a1 = MPZ(2*a + 3)
        if hyperbolic or a&1:
            return MPZ_ONE, a1 * q**2, a1
        else:
            return -MPZ_ONE, a1 * q**2, a1
    m = (a+b)//2
    p1, q1, r1 = bsp_acot(q, a, m, hyperbolic)
    p2, q2, r2 = bsp_acot(q, m, b, hyperbolic)
    return q2*p1 + r1*p2, q1*q2, r1*r2

# the acoth(x) series converges like the geometric series for x^2
# N = ceil(p*log(2)/(2*log(x)))
def acot_fixed(a, prec, hyperbolic):
    """
    Compute acot(a) or acoth(a) for an integer a with binary splitting; see
    http://numbers.computation.free.fr/Constants/Algorithms/splitting.html
    """
    N = int(0.35 * prec/math.log(a) + 20)
    p, q, r = bsp_acot(a, 0,N, hyperbolic)
    return ((p+q)<<prec)//(q*a)

def machin(coefs, prec, hyperbolic=False):
    """
    Evaluate a Machin-like formula, i.e., a linear combination of
    acot(n) or acoth(n) for specific integer values of n, using fixed-
    point arithmetic. The input should be a list [(c, n), ...], giving
    c*acot[h](n) + ...
    """
    extraprec = 10
    s = MPZ_ZERO
    for a, b in coefs:
        s += MPZ(a) * acot_fixed(MPZ(b), prec+extraprec, hyperbolic)
    return (s >> extraprec)

# Logarithms of integers are needed for various computations involving
# logarithms, powers, radix conversion, etc

@constant_memo
def ln2_fixed(prec):
    """
    Computes ln(2). This is done with a hyperbolic Machin-type formula,
    with binary splitting at high precision.
    """
    return machin([(18, 26), (-2, 4801), (8, 8749)], prec, True)

@constant_memo
def ln10_fixed(prec):
    """
    Computes ln(10). This is done with a hyperbolic Machin-type formula.
    """
    return machin([(46, 31), (34, 49), (20, 161)], prec, True)


"""
For computation of pi, we use the Chudnovsky series:

             oo
             ___        k
      1     \       (-1)  (6 k)! (A + B k)
    ----- =  )     -----------------------
    12 pi   /___               3  3k+3/2
                    (3 k)! (k!)  C
            k = 0

where A, B, and C are certain integer constants. This series adds roughly
14 digits per term. Note that C^(3/2) can be extracted so that the
series contains only rational terms. This makes binary splitting very
efficient.

The recurrence formulas for the binary splitting were taken from
ftp://ftp.gmplib.org/pub/src/gmp-chudnovsky.c

Previously, Machin's formula was used at low precision and the AGM iteration
was used at high precision. However, the Chudnovsky series is essentially as
fast as the Machin formula at low precision and in practice about 3x faster
than the AGM at high precision (despite theoretically having a worse
asymptotic complexity), so there is no reason not to use it in all cases.

"""

# Constants in Chudnovsky's series
CHUD_A = MPZ(13591409)
CHUD_B = MPZ(545140134)
CHUD_C = MPZ(640320)
CHUD_D = MPZ(12)

def bs_chudnovsky(a, b, level, verbose):
    """
    Computes the sum from a to b of the series in the Chudnovsky
    formula. Returns g, p, q where p/q is the sum as an exact
    fraction and g is a temporary value used to save work
    for recursive calls.
    """
    if b-a == 1:
        g = MPZ((6*b-5)*(2*b-1)*(6*b-1))
        p = b**3 * CHUD_C**3 // 24
        q = (-1)**b * g * (CHUD_A+CHUD_B*b)
    else:
        if verbose and level < 4:
            print("  binary splitting", a, b)
        mid = (a+b)//2
        g1, p1, q1 = bs_chudnovsky(a, mid, level+1, verbose)
        g2, p2, q2 = bs_chudnovsky(mid, b, level+1, verbose)
        p = p1*p2
        g = g1*g2
        q = q1*p2 + q2*g1
    return g, p, q

@constant_memo
def pi_fixed(prec, verbose=False, verbose_base=None):
    """
    Compute floor(pi * 2**prec) as a big integer.

    This is done using Chudnovsky's series (see comments in
    libelefun.py for details).
    """
    # The Chudnovsky series gives 14.18 digits per term
    N = int(prec/3.3219280948/14.181647462 + 2)
    if verbose:
        print("binary splitting with N =", N)
    g, p, q = bs_chudnovsky(0, N, 0, verbose)
    sqrtC = isqrt_fast(CHUD_C<<(2*prec))
    v = p*CHUD_C*sqrtC//((q+CHUD_A*p)*CHUD_D)
    return v

def degree_fixed(prec):
    return pi_fixed(prec)//180

def bspe(a, b):
    """
    Sum series for exp(1)-1 between a, b, returning the result
    as an exact fraction (p, q).
    """
    if b-a == 1:
        return MPZ_ONE, MPZ(b)
    m = (a+b)//2
    p1, q1 = bspe(a, m)
    p2, q2 = bspe(m, b)
    return p1*q2+p2, q1*q2

@constant_memo
def e_fixed(prec):
    """
    Computes exp(1). This is done using the ordinary Taylor series for
    exp, with binary splitting. For a description of the algorithm,
    see:

        http://numbers.computation.free.fr/Constants/
            Algorithms/splitting.html
    """
    # Slight overestimate of N needed for 1/N! < 2**(-prec)
    # This could be tightened for large N.
    N = int(1.1*prec/math.log(prec) + 20)
    p, q = bspe(0,N)
    return ((p+q)<<prec)//q

@constant_memo
def phi_fixed(prec):
    """
    Computes the golden ratio, (1+sqrt(5))/2
    """
    prec += 10
    a = isqrt_fast(MPZ_FIVE<<(2*prec)) + (MPZ_ONE << prec)
    return a >> 11

mpf_phi    = def_mpf_constant(phi_fixed)
mpf_pi     = def_mpf_constant(pi_fixed)
mpf_e      = def_mpf_constant(e_fixed)
mpf_degree = def_mpf_constant(degree_fixed)
mpf_ln2    = def_mpf_constant(ln2_fixed)
mpf_ln10   = def_mpf_constant(ln10_fixed)


@constant_memo
def ln_sqrt2pi_fixed(prec):
    wp = prec + 10
    # ln(sqrt(2*pi)) = ln(2*pi)/2
    return to_fixed(mpf_log(mpf_shift(mpf_pi(wp), 1), wp), prec-1)

@constant_memo
def sqrtpi_fixed(prec):
    return sqrt_fixed(pi_fixed(prec), prec)

mpf_sqrtpi   = def_mpf_constant(sqrtpi_fixed)
mpf_ln_sqrt2pi   = def_mpf_constant(ln_sqrt2pi_fixed)


#----------------------------------------------------------------------------#
#                                                                            #
#                                    Powers                                  #
#                                                                            #
#----------------------------------------------------------------------------#

def mpf_pow(s, t, prec, rnd=round_fast):
    """
    Compute s**t. Raises ComplexResult if s is negative and t is
    fractional.
    """
    ssign, sman, sexp, sbc = s
    tsign, tman, texp, tbc = t
    if ssign and texp < 0:
        raise ComplexResult("negative number raised to a fractional power")
    if texp >= 0:
        return mpf_pow_int(s, (-1)**tsign * (tman<<texp), prec, rnd)
    # s**(n/2) = sqrt(s)**n
    if texp == -1:
        if tman == 1:
            if tsign:
                return mpf_div(fone, mpf_sqrt(s, prec+10,
                    reciprocal_rnd[rnd]), prec, rnd)
            return mpf_sqrt(s, prec, rnd)
        else:
            if tsign:
                return mpf_pow_int(mpf_sqrt(s, prec+10,
                    reciprocal_rnd[rnd]), -tman, prec, rnd)
            return mpf_pow_int(mpf_sqrt(s, prec+10, rnd), tman, prec, rnd)
    # General formula: s**t = exp(t*log(s))
    # TODO: handle rnd direction of the logarithm carefully
    c = mpf_log(s, prec+10, rnd)
    return mpf_exp(mpf_mul(t, c), prec, rnd)

def int_pow_fixed(y, n, prec):
    """n-th power of a fixed point number with precision prec

       Returns the power in the form man, exp,
       man * 2**exp ~= y**n
    """
    if n == 2:
        return (y*y), 0
    bc = bitcount(y)
    exp = 0
    workprec = 2 * (prec + 4*bitcount(n) + 4)
    _, pm, pe, pbc = fone
    while 1:
        if n & 1:
            pm = pm*y
            pe = pe+exp
            pbc += bc - 2
            pbc = pbc + bctable[int(pm >> pbc)]
            if pbc > workprec:
                pm = pm >> (pbc-workprec)
                pe += pbc - workprec
                pbc = workprec
            n -= 1
            if not n:
                break
        y = y*y
        exp = exp+exp
        bc = bc + bc - 2
        bc = bc + bctable[int(y >> bc)]
        if bc > workprec:
            y = y >> (bc-workprec)
            exp += bc - workprec
            bc = workprec
        n = n // 2
    return pm, pe

# froot(s, n, prec, rnd) computes the real n-th root of a
# positive mpf tuple s.
# To compute the root we start from a 50-bit estimate for r
# generated with ordinary floating-point arithmetic, and then refine
# the value to full accuracy using the iteration

#            1  /                     y       \
#   r     = --- | (n-1)  * r   +  ----------  |
#    n+1     n  \           n     r_n**(n-1)  /

# which is simply Newton's method applied to the equation r**n = y.
# With giant_steps(start, prec+extra) = [p0,...,pm, prec+extra]
# and y = man * 2**-shift  one has
# (man * 2**exp)**(1/n) =
# y**(1/n) * 2**(start-prec/n) * 2**(p0-start) * ... * 2**(prec+extra-pm) *
# 2**((exp+shift-(n-1)*prec)/n -extra))
# The last factor is accounted for in the last line of froot.

def nthroot_fixed(y, n, prec, exp1):
    start = 50
    try:
        y1 = rshift(y, prec - n*start)
        r = MPZ(int(y1**(1.0/n)))
    except OverflowError:
        y1 = from_int(y1, start)
        fn = from_int(n)
        fn = mpf_rdiv_int(1, fn, start)
        r = mpf_pow(y1, fn, start)
        r = to_int(r)
    extra = 10
    extra1 = n
    prevp = start
    for p in giant_steps(start, prec+extra):
        pm, pe = int_pow_fixed(r, n-1, prevp)
        r2 = rshift(pm, (n-1)*prevp - p - pe - extra1)
        B = lshift(y, 2*p-prec+extra1)//r2
        r = (B + (n-1) * lshift(r, p-prevp))//n
        prevp = p
    return r

def mpf_nthroot(s, n, prec, rnd=round_fast):
    """nth-root of a positive number

    Use the Newton method when faster, otherwise use x**(1/n)
    """
    sign, man, exp, bc = s
    if sign:
        raise ComplexResult("nth root of a negative number")
    if not man:
        if s == fnan:
            return fnan
        if s == fzero:
            if n > 0:
                return fzero
            if n == 0:
                return fone
            return finf
        # Infinity
        if not n:
            return fnan
        if n < 0:
            return fzero
        return finf
    flag_inverse = False
    if n < 2:
        if n == 0:
            return fone
        if n == 1:
            return mpf_pos(s, prec, rnd)
        if n == -1:
            return mpf_div(fone, s, prec, rnd)
        # n < 0
        rnd = reciprocal_rnd[rnd]
        flag_inverse = True
        extra_inverse = 5
        prec += extra_inverse
        n = -n
    if n > 20 and (n >= 20000 or prec < int(233 + 28.3 * n**0.62)):
        prec2 = prec + 10
        fn = from_int(n)
        nth = mpf_rdiv_int(1, fn, prec2)
        r = mpf_pow(s, nth, prec2, rnd)
        s = normalize(r[0], r[1], r[2], r[3], prec, rnd)
        if flag_inverse:
            return mpf_div(fone, s, prec-extra_inverse, rnd)
        else:
            return s
    # Convert to a fixed-point number with prec2 bits.
    prec2 = prec + 2*n - (prec%n)
    # a few tests indicate that
    # for 10 < n < 10**4 a bit more precision is needed
    if n > 10:
        prec2 += prec2//10
        prec2 = prec2 - prec2%n
    # Mantissa may have more bits than we need. Trim it down.
    shift = bc - prec2
    # Adjust exponents to make prec2 and exp+shift multiples of n.
    sign1 = 0
    es = exp+shift
    if es < 0:
        sign1 = 1
        es = -es
    if sign1:
        shift += es%n
    else:
        shift -= es%n
    man = rshift(man, shift)
    extra = 10
    exp1 = ((exp+shift-(n-1)*prec2)//n) - extra
    rnd_shift = 0
    if flag_inverse:
        if rnd == 'u' or rnd == 'c':
            rnd_shift = 1
    else:
        if rnd == 'd' or rnd == 'f':
            rnd_shift = 1
    man = nthroot_fixed(man+rnd_shift, n, prec2, exp1)
    s = from_man_exp(man, exp1, prec, rnd)
    if flag_inverse:
        return mpf_div(fone, s, prec-extra_inverse, rnd)
    else:
        return s

def mpf_cbrt(s, prec, rnd=round_fast):
    """cubic root of a positive number"""
    return mpf_nthroot(s, 3, prec, rnd)

#----------------------------------------------------------------------------#
#                                                                            #
#                                Logarithms                                  #
#                                                                            #
#----------------------------------------------------------------------------#


def log_int_fixed(n, prec, ln2=None):
    """
    Fast computation of log(n), caching the value for small n,
    intended for zeta sums.
    """
    if n in log_int_cache:
        value, vprec = log_int_cache[n]
        if vprec >= prec:
            return value >> (vprec - prec)
    wp = prec + 10
    if wp <= LOG_TAYLOR_SHIFT:
        if ln2 is None:
            ln2 = ln2_fixed(wp)
        r = bitcount(n)
        x = n << (wp-r)
        v = log_taylor_cached(x, wp) + r*ln2
    else:
        v = to_fixed(mpf_log(from_int(n), wp+5), wp)
    if n < MAX_LOG_INT_CACHE:
        log_int_cache[n] = (v, wp)
    return v >> (wp-prec)

def agm_fixed(a, b, prec):
    """
    Fixed-point computation of agm(a,b), assuming
    a, b both close to unit magnitude.
    """
    i = 0
    while 1:
        anew = (a+b)>>1
        if i > 4 and abs(a-anew) < 8:
            return a
        b = isqrt_fast(a*b)
        a = anew
        i += 1
    return a

def log_agm(x, prec):
    """
    Fixed-point computation of -log(x) = log(1/x), suitable
    for large precision. It is required that 0 < x < 1. The
    algorithm used is the Sasaki-Kanada formula

        -log(x) = pi/agm(theta2(x)^2,theta3(x)^2). [1]

    For faster convergence in the theta functions, x should
    be chosen closer to 0.

    Guard bits must be added by the caller.

    HYPOTHESIS: if x = 2^(-n), n bits need to be added to
    account for the truncation to a fixed-point number,
    and this is the only significant cancellation error.

    The number of bits lost to roundoff is small and can be
    considered constant.

    [1] Richard P. Brent, "Fast Algorithms for High-Precision
        Computation of Elementary Functions (extended abstract)",
        http://wwwmaths.anu.edu.au/~brent/pd/RNC7-Brent.pdf

    """
    x2 = (x*x) >> prec
    # Compute jtheta2(x)**2
    s = a = b = x2
    while a:
        b = (b*x2) >> prec
        a = (a*b) >> prec
        s += a
    s += (MPZ_ONE<<prec)
    s = (s*s)>>(prec-2)
    s = (s*isqrt_fast(x<<prec))>>prec
    # Compute jtheta3(x)**2
    t = a = b = x
    while a:
        b = (b*x2) >> prec
        a = (a*b) >> prec
        t += a
    t = (MPZ_ONE<<prec) + (t<<1)
    t = (t*t)>>prec
    # Final formula
    p = agm_fixed(s, t, prec)
    return (pi_fixed(prec) << prec) // p

def log_taylor(x, prec, r=0):
    """
    Fixed-point calculation of log(x). It is assumed that x is close
    enough to 1 for the Taylor series to converge quickly. Convergence
    can be improved by specifying r > 0 to compute
    log(x^(1/2^r))*2^r, at the cost of performing r square roots.

    The caller must provide sufficient guard bits.
    """
    for i in xrange(r):
        x = isqrt_fast(x<<prec)
    one = MPZ_ONE << prec
    v = ((x-one)<<prec)//(x+one)
    sign = v < 0
    if sign:
        v = -v
    v2 = (v*v) >> prec
    v4 = (v2*v2) >> prec
    s0 = v
    s1 = v//3
    v = (v*v4) >> prec
    k = 5
    while v:
        s0 += v // k
        k += 2
        s1 += v // k
        v = (v*v4) >> prec
        k += 2
    s1 = (s1*v2) >> prec
    s = (s0+s1) << (1+r)
    if sign:
        return -s
    return s

def log_taylor_cached(x, prec):
    """
    Fixed-point computation of log(x), assuming x in (0.5, 2)
    and prec <= LOG_TAYLOR_PREC.
    """
    n = x >> (prec-LOG_TAYLOR_SHIFT)
    cached_prec = cache_prec_steps[prec]
    dprec = cached_prec - prec
    if (n, cached_prec) in log_taylor_cache:
        a, log_a = log_taylor_cache[n, cached_prec]
    else:
        a = n << (cached_prec - LOG_TAYLOR_SHIFT)
        log_a = log_taylor(a, cached_prec, 8)
        log_taylor_cache[n, cached_prec] = (a, log_a)
    a >>= dprec
    log_a >>= dprec
    u = ((x - a) << prec) // a
    v = (u << prec) // ((MPZ_TWO << prec) + u)
    v2 = (v*v) >> prec
    v4 = (v2*v2) >> prec
    s0 = v
    s1 = v//3
    v = (v*v4) >> prec
    k = 5
    while v:
        s0 += v//k
        k += 2
        s1 += v//k
        v = (v*v4) >> prec
        k += 2
    s1 = (s1*v2) >> prec
    s = (s0+s1) << 1
    return log_a + s

def mpf_log(x, prec, rnd=round_fast):
    """
    Compute the natural logarithm of the mpf value x. If x is negative,
    ComplexResult is raised.
    """
    sign, man, exp, bc = x
    #------------------------------------------------------------------
    # Handle special values
    if not man:
        if x == fzero: return fninf
        if x == finf: return finf
        if x == fnan: return fnan
    if sign:
        raise ComplexResult("logarithm of a negative number")
    wp = prec + 20
    #------------------------------------------------------------------
    # Handle log(2^n) = log(n)*2.
    # Here we catch the only possible exact value, log(1) = 0
    if man == 1:
        if not exp:
            return fzero
        return from_man_exp(exp*ln2_fixed(wp), -wp, prec, rnd)
    mag = exp+bc
    abs_mag = abs(mag)
    #------------------------------------------------------------------
    # Handle x = 1+eps, where log(x) ~ x. We need to check for
    # cancellation when moving to fixed-point math and compensate
    # by increasing the precision. Note that abs_mag in (0, 1) <=>
    # 0.5 < x < 2 and x != 1
    if abs_mag <= 1:
        # Calculate t = x-1 to measure distance from 1 in bits
        tsign = 1-abs_mag
        if tsign:
            tman = (MPZ_ONE<<bc) - man
        else:
            tman = man - (MPZ_ONE<<(bc-1))
        tbc = bitcount(tman)
        cancellation = bc - tbc
        if cancellation > wp:
            t = normalize(tsign, tman, abs_mag-bc, tbc, tbc, 'n')
            return mpf_perturb(t, tsign, prec, rnd)
        else:
            wp += cancellation
        # TODO: if close enough to 1, we could use Taylor series
        # even in the AGM precision range, since the Taylor series
        # converges rapidly
    #------------------------------------------------------------------
    # Another special case:
    # n*log(2) is a good enough approximation
    if abs_mag > 10000:
        if bitcount(abs_mag) > wp:
            return from_man_exp(exp*ln2_fixed(wp), -wp, prec, rnd)
    #------------------------------------------------------------------
    # General case.
    # Perform argument reduction using log(x) = log(x*2^n) - n*log(2):
    # If we are in the Taylor precision range, choose magnitude 0 or 1.
    # If we are in the AGM precision range, choose magnitude -m for
    # some large m; benchmarking on one machine showed m = prec/20 to be
    # optimal between 1000 and 100,000 digits.
    if wp <= LOG_TAYLOR_PREC:
        m = log_taylor_cached(lshift(man, wp-bc), wp)
        if mag:
            m += mag*ln2_fixed(wp)
    else:
        optimal_mag = -wp//LOG_AGM_MAG_PREC_RATIO
        n = optimal_mag - mag
        x = mpf_shift(x, n)
        wp += (-optimal_mag)
        m = -log_agm(to_fixed(x, wp), wp)
        m -= n*ln2_fixed(wp)
    return from_man_exp(m, -wp, prec, rnd)

def mpf_log_hypot(a, b, prec, rnd):
    """
    Computes log(sqrt(a^2+b^2)) accurately.
    """
    # If either a or b is inf/nan/0, assume it to be a
    if not b[1]:
        a, b = b, a
    # a is inf/nan/0
    if not a[1]:
        # both are inf/nan/0
        if not b[1]:
            if a == b == fzero:
                return fninf
            if fnan in (a, b):
                return fnan
            # at least one term is (+/- inf)^2
            return finf
        # only a is inf/nan/0
        if a == fzero:
            # log(sqrt(0+b^2)) = log(|b|)
            return mpf_log(mpf_abs(b), prec, rnd)
        if a == fnan:
            return fnan
        return finf
    # Exact
    a2 = mpf_mul(a,a)
    b2 = mpf_mul(b,b)
    extra = 20
    # Not exact
    h2 = mpf_add(a2, b2, prec+extra)
    cancelled = mpf_add(h2, fnone, 10)
    mag_cancelled = cancelled[2]+cancelled[3]
    # Just redo the sum exactly if necessary (could be smarter
    # and avoid memory allocation when a or b is precisely 1
    # and the other is tiny...)
    if cancelled == fzero or mag_cancelled < -extra//2:
        h2 = mpf_add(a2, b2, prec+extra-min(a2[2],b2[2]))
    return mpf_shift(mpf_log(h2, prec, rnd), -1)


#----------------------------------------------------------------------
# Inverse tangent
#

def atan_newton(x, prec):
    if prec >= 100:
        r = math.atan(int((x>>(prec-53)))/2.0**53)
    else:
        r = math.atan(int(x)/2.0**prec)
    prevp = 50
    r = MPZ(int(r * 2.0**53) >> (53-prevp))
    extra_p = 50
    for wp in giant_steps(prevp, prec):
        wp += extra_p
        r = r << (wp-prevp)
        cos, sin = cos_sin_fixed(r, wp)
        tan = (sin << wp) // cos
        a = ((tan-rshift(x, prec-wp)) << wp) // ((MPZ_ONE<<wp) + ((tan**2)>>wp))
        r = r - a
        prevp = wp
    return rshift(r, prevp-prec)

def atan_taylor_get_cached(n, prec):
    # Taylor series with caching wins up to huge precisions
    # To avoid unnecessary precomputation at low precision, we
    # do it in steps
    # Round to next power of 2
    prec2 = (1<<(bitcount(prec-1))) + 20
    dprec = prec2 - prec
    if (n, prec2) in atan_taylor_cache:
        a, atan_a = atan_taylor_cache[n, prec2]
    else:
        a = n << (prec2 - ATAN_TAYLOR_SHIFT)
        atan_a = atan_newton(a, prec2)
        atan_taylor_cache[n, prec2] = (a, atan_a)
    return (a >> dprec), (atan_a >> dprec)

def atan_taylor(x, prec):
    n = (x >> (prec-ATAN_TAYLOR_SHIFT))
    a, atan_a = atan_taylor_get_cached(n, prec)
    d = x - a
    s0 = v = (d << prec) // ((a**2 >> prec) + (a*d >> prec) + (MPZ_ONE << prec))
    v2 = (v**2 >> prec)
    v4 = (v2 * v2) >> prec
    s1 = v//3
    v = (v * v4) >> prec
    k = 5
    while v:
        s0 += v // k
        k += 2
        s1 += v // k
        v = (v * v4) >> prec
        k += 2
    s1 = (s1 * v2) >> prec
    s = s0 - s1
    return atan_a + s

def atan_inf(sign, prec, rnd):
    if not sign:
        return mpf_shift(mpf_pi(prec, rnd), -1)
    return mpf_neg(mpf_shift(mpf_pi(prec, negative_rnd[rnd]), -1))

def mpf_atan(x, prec, rnd=round_fast):
    sign, man, exp, bc = x
    if not man:
        if x == fzero: return fzero
        if x == finf: return atan_inf(0, prec, rnd)
        if x == fninf: return atan_inf(1, prec, rnd)
        return fnan
    mag = exp + bc
    # Essentially infinity
    if mag > prec+20:
        return atan_inf(sign, prec, rnd)
    # Essentially ~ x
    if -mag > prec+20:
        return mpf_perturb(x, 1-sign, prec, rnd)
    wp = prec + 30 + abs(mag)
    # For large x, use atan(x) = pi/2 - atan(1/x)
    if mag >= 2:
        x = mpf_rdiv_int(1, x, wp)
        reciprocal = True
    else:
        reciprocal = False
    t = to_fixed(x, wp)
    if sign:
        t = -t
    if wp < ATAN_TAYLOR_PREC:
        a = atan_taylor(t, wp)
    else:
        a = atan_newton(t, wp)
    if reciprocal:
        a = ((pi_fixed(wp)>>1)+1) - a
    if sign:
        a = -a
    return from_man_exp(a, -wp, prec, rnd)

# TODO: cleanup the special cases
def mpf_atan2(y, x, prec, rnd=round_fast):
    xsign, xman, xexp, xbc = x
    ysign, yman, yexp, ybc = y
    if not yman:
        if y == fzero and x != fnan:
            if mpf_sign(x) >= 0:
                return fzero
            return mpf_pi(prec, rnd)
        if y in (finf, fninf):
            if x in (finf, fninf):
                return fnan
            # pi/2
            if y == finf:
                return mpf_shift(mpf_pi(prec, rnd), -1)
            # -pi/2
            return mpf_neg(mpf_shift(mpf_pi(prec, negative_rnd[rnd]), -1))
        return fnan
    if ysign:
        return mpf_neg(mpf_atan2(mpf_neg(y), x, prec, negative_rnd[rnd]))
    if not xman:
        if x == fnan:
            return fnan
        if x == finf:
            return fzero
        if x == fninf:
            return mpf_pi(prec, rnd)
        if y == fzero:
            return fzero
        return mpf_shift(mpf_pi(prec, rnd), -1)
    tquo = mpf_atan(mpf_div(y, x, prec+4), prec+4)
    if xsign:
        return mpf_add(mpf_pi(prec+4), tquo, prec, rnd)
    else:
        return mpf_pos(tquo, prec, rnd)

def mpf_asin(x, prec, rnd=round_fast):
  sign, man, exp, bc = x
  if bc+exp > 0 and x not in (fone, fnone):
      raise ComplexResult("asin(x) is real only for -1 <= x <= 1")
  # asin(x) = 2*atan(x/(1+sqrt(1-x**2)))
  wp = prec + 15
  a = mpf_mul(x, x)
  b = mpf_add(fone, mpf_sqrt(mpf_sub(fone, a, wp), wp), wp)
  c = mpf_div(x, b, wp)
  return mpf_shift(mpf_atan(c, prec, rnd), 1)

def mpf_acos(x, prec, rnd=round_fast):
    # acos(x) = 2*atan(sqrt(1-x**2)/(1+x))
    sign, man, exp, bc = x
    if bc + exp > 0:
        if x not in (fone, fnone):
            raise ComplexResult("acos(x) is real only for -1 <= x <= 1")
        if x == fnone:
            return mpf_pi(prec, rnd)
    wp = prec + 15
    a = mpf_mul(x, x)
    b = mpf_sqrt(mpf_sub(fone, a, wp), wp)
    c = mpf_div(b, mpf_add(fone, x, wp), wp)
    return mpf_shift(mpf_atan(c, prec, rnd), 1)

def mpf_asinh(x, prec, rnd=round_fast):
    wp = prec + 20
    sign, man, exp, bc = x
    mag = exp+bc
    if mag < -8:
        if mag < -wp:
            return mpf_perturb(x, 1-sign, prec, rnd)
        wp += (-mag)
    # asinh(x) = log(x+sqrt(x**2+1))
    # use reflection symmetry to avoid cancellation
    q = mpf_sqrt(mpf_add(mpf_mul(x, x), fone, wp), wp)
    q = mpf_add(mpf_abs(x), q, wp)
    if sign:
        return mpf_neg(mpf_log(q, prec, negative_rnd[rnd]))
    else:
        return mpf_log(q, prec, rnd)

def mpf_acosh(x, prec, rnd=round_fast):
    # acosh(x) = log(x+sqrt(x**2-1))
    wp = prec + 15
    if mpf_cmp(x, fone) == -1:
        raise ComplexResult("acosh(x) is real only for x >= 1")
    q = mpf_sqrt(mpf_add(mpf_mul(x,x), fnone, wp), wp)
    return mpf_log(mpf_add(x, q, wp), prec, rnd)

def mpf_atanh(x, prec, rnd=round_fast):
    # atanh(x) = log((1+x)/(1-x))/2
    sign, man, exp, bc = x
    if (not man) and exp:
        if x in (fzero, fnan):
            return x
        raise ComplexResult("atanh(x) is real only for -1 <= x <= 1")
    mag = bc + exp
    if mag > 0:
        if mag == 1 and man == 1:
            return [finf, fninf][sign]
        raise ComplexResult("atanh(x) is real only for -1 <= x <= 1")
    wp = prec + 15
    if mag < -8:
        if mag < -wp:
            return mpf_perturb(x, sign, prec, rnd)
        wp += (-mag)
    a = mpf_add(x, fone, wp)
    b = mpf_sub(fone, x, wp)
    return mpf_shift(mpf_log(mpf_div(a, b, wp), prec, rnd), -1)

def mpf_fibonacci(x, prec, rnd=round_fast):
    sign, man, exp, bc = x
    if not man:
        if x == fninf:
            return fnan
        return x
    # F(2^n) ~= 2^(2^n)
    size = abs(exp+bc)
    if exp >= 0:
        # Exact
        if size < 10 or size <= bitcount(prec):
            return from_int(ifib(to_int(x)), prec, rnd)
    # Use the modified Binet formula
    wp = prec + size + 20
    a = mpf_phi(wp)
    b = mpf_add(mpf_shift(a, 1), fnone, wp)
    u = mpf_pow(a, x, wp)
    v = mpf_cos_pi(x, wp)
    v = mpf_div(v, u, wp)
    u = mpf_sub(u, v, wp)
    u = mpf_div(u, b, prec, rnd)
    return u


#-------------------------------------------------------------------------------
# Exponential-type functions
#-------------------------------------------------------------------------------

def exponential_series(x, prec, type=0):
    """
    Taylor series for cosh/sinh or cos/sin.

    type = 0 -- returns exp(x)  (slightly faster than cosh+sinh)
    type = 1 -- returns (cosh(x), sinh(x))
    type = 2 -- returns (cos(x), sin(x))
    """
    if x < 0:
        x = -x
        sign = 1
    else:
        sign = 0
    r = int(0.5*prec**0.5)
    xmag = bitcount(x) - prec
    r = max(0, xmag + r)
    extra = 10 + 2*max(r,-xmag)
    wp = prec + extra
    x <<= (extra - r)
    one = MPZ_ONE << wp
    alt = (type == 2)
    if prec < EXP_SERIES_U_CUTOFF:
        x2 = a = (x*x) >> wp
        x4 = (x2*x2) >> wp
        s0 = s1 = MPZ_ZERO
        k = 2
        while a:
            a //= (k-1)*k; s0 += a; k += 2
            a //= (k-1)*k; s1 += a; k += 2
            a = (a*x4) >> wp
        s1 = (x2*s1) >> wp
        if alt:
            c = s1 - s0 + one
        else:
            c = s1 + s0 + one
    else:
        u = int(0.3*prec**0.35)
        x2 = a = (x*x) >> wp
        xpowers = [one, x2]
        for i in xrange(1, u):
            xpowers.append((xpowers[-1]*x2)>>wp)
        sums = [MPZ_ZERO] * u
        k = 2
        while a:
            for i in xrange(u):
                a //= (k-1)*k
                if alt and k & 2: sums[i] -= a
                else:             sums[i] += a
                k += 2
            a = (a*xpowers[-1]) >> wp
        for i in xrange(1, u):
            sums[i] = (sums[i]*xpowers[i]) >> wp
        c = sum(sums) + one
    if type == 0:
        s = isqrt_fast(c*c - (one<<wp))
        if sign:
            v = c - s
        else:
            v = c + s
        for i in xrange(r):
            v = (v*v) >> wp
        return v >> extra
    else:
        # Repeatedly apply the double-angle formula
        # cosh(2*x) = 2*cosh(x)^2 - 1
        # cos(2*x) = 2*cos(x)^2 - 1
        pshift = wp-1
        for i in xrange(r):
            c = ((c*c) >> pshift) - one
        # With the abs, this is the same for sinh and sin
        s = isqrt_fast(abs((one<<wp) - c*c))
        if sign:
            s = -s
        return (c>>extra), (s>>extra)

def exp_basecase(x, prec):
    """
    Compute exp(x) as a fixed-point number. Works for any x,
    but for speed should have |x| < 1. For an arbitrary number,
    use exp(x) = exp(x-m*log(2)) * 2^m where m = floor(x/log(2)).
    """
    if prec > EXP_COSH_CUTOFF:
        return exponential_series(x, prec, 0)
    r = int(prec**0.5)
    prec += r
    s0 = s1 = (MPZ_ONE << prec)
    k = 2
    a = x2 = (x*x) >> prec
    while a:
        a //= k; s0 += a; k += 1
        a //= k; s1 += a; k += 1
        a = (a*x2) >> prec
    s1 = (s1*x) >> prec
    s = s0 + s1
    u = r
    while r:
        s = (s*s) >> prec
        r -= 1
    return s >> u

def exp_expneg_basecase(x, prec):
    """
    Computation of exp(x), exp(-x)
    """
    if prec > EXP_COSH_CUTOFF:
        cosh, sinh = exponential_series(x, prec, 1)
        return cosh+sinh, cosh-sinh
    a = exp_basecase(x, prec)
    b = (MPZ_ONE << (prec+prec)) // a
    return a, b

def cos_sin_basecase(x, prec):
    """
    Compute cos(x), sin(x) as fixed-point numbers, assuming x
    in [0, pi/2). For an arbitrary number, use x' = x - m*(pi/2)
    where m = floor(x/(pi/2)) along with quarter-period symmetries.
    """
    if prec > COS_SIN_CACHE_PREC:
        return exponential_series(x, prec, 2)
    precs = prec - COS_SIN_CACHE_STEP
    t = x >> precs
    n = int(t)
    if n not in cos_sin_cache:
        w = t<<(10+COS_SIN_CACHE_PREC-COS_SIN_CACHE_STEP)
        cos_t, sin_t = exponential_series(w, 10+COS_SIN_CACHE_PREC, 2)
        cos_sin_cache[n] = (cos_t>>10), (sin_t>>10)
    cos_t, sin_t = cos_sin_cache[n]
    offset = COS_SIN_CACHE_PREC - prec
    cos_t >>= offset
    sin_t >>= offset
    x -= t << precs
    cos = MPZ_ONE << prec
    sin = x
    k = 2
    a = -((x*x) >> prec)
    while a:
        a //= k; cos += a; k += 1; a = (a*x) >> prec
        a //= k; sin += a; k += 1; a = -((a*x) >> prec)
    return ((cos*cos_t-sin*sin_t) >> prec), ((sin*cos_t+cos*sin_t) >> prec)

def mpf_exp(x, prec, rnd=round_fast):
    sign, man, exp, bc = x
    if man:
        mag = bc + exp
        wp = prec + 14
        if sign:
            man = -man
        # TODO: the best cutoff depends on both x and the precision.
        if prec > 600 and exp >= 0:
            # Need about log2(exp(n)) ~= 1.45*mag extra precision
            e = mpf_e(wp+int(1.45*mag))
            return mpf_pow_int(e, man<<exp, prec, rnd)
        if mag < -wp:
            return mpf_perturb(fone, sign, prec, rnd)
        # |x| >= 2
        if mag > 1:
            # For large arguments: exp(2^mag*(1+eps)) =
            # exp(2^mag)*exp(2^mag*eps) = exp(2^mag)*(1 + 2^mag*eps + ...)
            # so about mag extra bits is required.
            wpmod = wp + mag
            offset = exp + wpmod
            if offset >= 0:
                t = man << offset
            else:
                t = man >> (-offset)
            lg2 = ln2_fixed(wpmod)
            n, t = divmod(t, lg2)
            n = int(n)
            t >>= mag
        else:
            offset = exp + wp
            if offset >= 0:
                t = man << offset
            else:
                t = man >> (-offset)
            n = 0
        man = exp_basecase(t, wp)
        return from_man_exp(man, n-wp, prec, rnd)
    if not exp:
        return fone
    if x == fninf:
        return fzero
    return x


def mpf_cosh_sinh(x, prec, rnd=round_fast, tanh=0):
    """Simultaneously compute (cosh(x), sinh(x)) for real x"""
    sign, man, exp, bc = x
    if (not man) and exp:
        if tanh:
            if x == finf: return fone
            if x == fninf: return fnone
            return fnan
        if x == finf: return (finf, finf)
        if x == fninf: return (finf, fninf)
        return fnan, fnan
    mag = exp+bc
    wp = prec+14
    if mag < -4:
        # Extremely close to 0, sinh(x) ~= x and cosh(x) ~= 1
        if mag < -wp:
            if tanh:
                return mpf_perturb(x, 1-sign, prec, rnd)
            cosh = mpf_perturb(fone, 0, prec, rnd)
            sinh = mpf_perturb(x, sign, prec, rnd)
            return cosh, sinh
        # Fix for cancellation when computing sinh
        wp += (-mag)
    # Does exp(-2*x) vanish?
    if mag > 10:
        if 3*(1<<(mag-1)) > wp:
            # XXX: rounding
            if tanh:
                return mpf_perturb([fone,fnone][sign], 1-sign, prec, rnd)
            c = s = mpf_shift(mpf_exp(mpf_abs(x), prec, rnd), -1)
            if sign:
                s = mpf_neg(s)
            return c, s
    # |x| > 1
    if mag > 1:
        wpmod = wp + mag
        offset = exp + wpmod
        if offset >= 0:
            t = man << offset
        else:
            t = man >> (-offset)
        lg2 = ln2_fixed(wpmod)
        n, t = divmod(t, lg2)
        n = int(n)
        t >>= mag
    else:
        offset = exp + wp
        if offset >= 0:
            t = man << offset
        else:
            t = man >> (-offset)
        n = 0
    a, b = exp_expneg_basecase(t, wp)
    # TODO: optimize division precision
    cosh = a + (b>>(2*n))
    sinh = a - (b>>(2*n))
    if sign:
        sinh = -sinh
    if tanh:
        man = (sinh << wp) // cosh
        return from_man_exp(man, -wp, prec, rnd)
    else:
        cosh = from_man_exp(cosh, n-wp-1, prec, rnd)
        sinh = from_man_exp(sinh, n-wp-1, prec, rnd)
        return cosh, sinh


def mod_pi2(man, exp, mag, wp):
    # Reduce to standard interval
    if mag > 0:
        i = 0
        while 1:
            cancellation_prec = 20 << i
            wpmod = wp + mag + cancellation_prec
            pi2 = pi_fixed(wpmod-1)
            pi4 = pi2 >> 1
            offset = wpmod + exp
            if offset >= 0:
                t = man << offset
            else:
                t = man >> (-offset)
            n, y = divmod(t, pi2)
            if y > pi4:
                small = pi2 - y
            else:
                small = y
            if small >> (wp+mag-10):
                n = int(n)
                t = y >> mag
                wp = wpmod - mag
                break
            i += 1
    else:
        wp += (-mag)
        offset = exp + wp
        if offset >= 0:
            t = man << offset
        else:
            t = man >> (-offset)
        n = 0
    return t, n, wp


def mpf_cos_sin(x, prec, rnd=round_fast, which=0, pi=False):
    """
    which:
    0 -- return cos(x), sin(x)
    1 -- return cos(x)
    2 -- return sin(x)
    3 -- return tan(x)

    if pi=True, compute for pi*x
    """
    sign, man, exp, bc = x
    if not man:
        if exp:
            c, s = fnan, fnan
        else:
            c, s = fone, fzero
        if which == 0: return c, s
        if which == 1: return c
        if which == 2: return s
        if which == 3: return s

    mag = bc + exp
    wp = prec + 10

    # Extremely small?
    if mag < 0:
        if mag < -wp:
            if pi:
                x = mpf_mul(x, mpf_pi(wp))
            c = mpf_perturb(fone, 1, prec, rnd)
            s = mpf_perturb(x, 1-sign, prec, rnd)
            if which == 0: return c, s
            if which == 1: return c
            if which == 2: return s
            if which == 3: return mpf_perturb(x, sign, prec, rnd)
    if pi:
        if exp >= -1:
            if exp == -1:
                c = fzero
                s = (fone, fnone)[bool(man & 2) ^ sign]
            elif exp == 0:
                c, s = (fnone, fzero)
            else:
                c, s = (fone, fzero)
            if which == 0: return c, s
            if which == 1: return c
            if which == 2: return s
            if which == 3: return mpf_div(s, c, prec, rnd)
        # Subtract nearest half-integer (= mod by pi/2)
        n = ((man >> (-exp-2)) + 1) >> 1
        man = man - (n << (-exp-1))
        mag2 = bitcount(man) + exp
        wp = prec + 10 - mag2
        offset = exp + wp
        if offset >= 0:
            t = man << offset
        else:
            t = man >> (-offset)
        t = (t*pi_fixed(wp)) >> wp
    else:
        t, n, wp = mod_pi2(man, exp, mag, wp)
    c, s = cos_sin_basecase(t, wp)
    m = n & 3
    if   m == 1: c, s = -s, c
    elif m == 2: c, s = -c, -s
    elif m == 3: c, s = s, -c
    if sign:
        s = -s
    if which == 0:
        c = from_man_exp(c, -wp, prec, rnd)
        s = from_man_exp(s, -wp, prec, rnd)
        return c, s
    if which == 1:
        return from_man_exp(c, -wp, prec, rnd)
    if which == 2:
        return from_man_exp(s, -wp, prec, rnd)
    if which == 3:
        return from_rational(s, c, prec, rnd)

def mpf_cos(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 1)
def mpf_sin(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 2)
def mpf_tan(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 3)
def mpf_cos_sin_pi(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 0, 1)
def mpf_cos_pi(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 1, 1)
def mpf_sin_pi(x, prec, rnd=round_fast): return mpf_cos_sin(x, prec, rnd, 2, 1)
def mpf_cosh(x, prec, rnd=round_fast): return mpf_cosh_sinh(x, prec, rnd)[0]
def mpf_sinh(x, prec, rnd=round_fast): return mpf_cosh_sinh(x, prec, rnd)[1]
def mpf_tanh(x, prec, rnd=round_fast): return mpf_cosh_sinh(x, prec, rnd, tanh=1)


# Low-overhead fixed-point versions

def cos_sin_fixed(x, prec, pi2=None):
    if pi2 is None:
        pi2 = pi_fixed(prec-1)
    n, t = divmod(x, pi2)
    n = int(n)
    c, s = cos_sin_basecase(t, prec)
    m = n & 3
    if m == 0: return c, s
    if m == 1: return -s, c
    if m == 2: return -c, -s
    if m == 3: return s, -c

def exp_fixed(x, prec, ln2=None):
    if ln2 is None:
        ln2 = ln2_fixed(prec)
    n, t = divmod(x, ln2)
    n = int(n)
    v = exp_basecase(t, prec)
    if n >= 0:
        return v << n
    else:
        return v >> (-n)


if BACKEND == 'sage':
    try:
        import sage.libs.mpmath.ext_libmp as _lbmp
        mpf_sqrt = _lbmp.mpf_sqrt
        mpf_exp = _lbmp.mpf_exp
        mpf_log = _lbmp.mpf_log
        mpf_cos = _lbmp.mpf_cos
        mpf_sin = _lbmp.mpf_sin
        mpf_pow = _lbmp.mpf_pow
        exp_fixed = _lbmp.exp_fixed
        cos_sin_fixed = _lbmp.cos_sin_fixed
        log_int_fixed = _lbmp.log_int_fixed
    except (ImportError, AttributeError):
        print("Warning: Sage imports in libelefun failed")



from sympycore.arithmetic.numbers import *
from sympycore.arithmetic.infinity import *

mpq = normalized_fraction
mpc = mpqc

nan = undefined = Infinity(0)
oo = Infinity(1)
moo = Infinity(-1)
zoo = Infinity(undefined)

def test_Fraction():
    assert mpq(1) == 1
    assert mpq(1,1) == 1
    assert mpq(1,2) != 1
    assert mpq(1,2) == mpq(2,4)
    assert mpq(1,2) + mpq(5,6) == mpq(4,3)
    assert 2*mpq(1,2) == 1
    assert mpq(1,2)*2 == 1
    assert mpq(1,3) + mpq(2,3) == 1
    assert float(mpq(1,4)) == 0.25
    assert mpq(1,2) == mpq(-1,-2)
    assert mpq(-1,2) == mpq(1,-2)
    assert -mpq(2,3) == mpq(-2,3)
    assert +mpq(2,3) == mpq(2,3)
    assert mpq(1,2) - 1 == mpq(-1,2)
    assert 1 - mpq(1,2) == mpq(1,2)
    assert mpq(2,3)**0 == 1
    assert mpq(2,3)**1 == mpq(2,3)
    assert mpq(2,3)**2 == mpq(4,9)
    assert mpq(-2,3)**2 == mpq(4,9)
    assert mpq(-2,3)**-2 == mpq(9,4)
    assert mpq(-2,3)**3 == mpq(-8,27)
    assert mpq(-2,3)**-3 == mpq(-27,8)
    assert div(1,2) == mpq(1,2)
    assert div(3,mpq(1,2)) == 6
    assert div(mpq(1,2),mpq(3,2)) == mpq(1,3)
    assert mpq(1234,15) < 83
    assert mpq(1234,15) > 82
    assert mpq(2,3) < mpq(3, 4)
    assert mpq(28,3) // 4 == 2
    assert mpq(-28,3) // 4 == -3
    assert 4 // mpq(7,3) == 1
    assert 4 // mpq(-7,3) == -2
    assert mpq(5,4) % 1 == mpq(1,4)
    assert mpq(-5,4) % 1 == mpq(3,4)
    assert mpq(17,4) % mpq(6,7) == mpq(23,28)
    assert mpq(-17,4) % mpq(6,7) == mpq(1,28)

def test_mpf():
    # XXX
    setdps(15)
    assert mpf(2) != 3
    assert mpf(2) == 2
    assert mpf(1.1) == mpf('1.1')
    assert mpf(3) * mpf(4) == 12
    assert mpf(3) + mpf(4) == mpf(7)
    assert mpf(3) - mpf(4) == -1
    assert -mpf(3) == -3
    assert 4*mpf(3) == 12
    assert mpf(3)*4 == 12
    assert 2+mpf(5) == 7
    assert 0.5 + mpf(1) == 1.5
    assert mpf(1) + 0.5 == 1.5
    assert hash(mpf(3)) == hash(3)
    assert hash(mpf(1.5)) == hash(1.5)
    assert mpf(3) < mpf(4)
    assert mpf(3) > mpf(2)
    assert mpf(3) < 4

def test_mpc():
    assert mpc(2,3) == mpc(2,3)
    assert mpc(2,3) != mpc(2,4)
    assert mpc(3,0) == 3
    assert mpc(3,1) != 3
    assert mpc(2,3)*2 == mpc(4,6)
    assert 2*mpc(2,3) == mpc(4,6)
    assert mpc(2,3)+2 == mpc(4,3)
    assert 2+mpc(2,3) == mpc(4,3)
    assert mpc(2,3)-2 == mpc(0,3)
    assert 2-mpc(2,3) == mpc(0,-3)
    assert mpc(mpq(2,3), 1) + mpc(mpq(1,3), 1) == mpc(1, 2)
    assert -mpc(2,3) == mpc(-2,-3)
    assert +mpc(2,3) == mpc(2,3)
    assert mpc(0,1)**2 == -1
    assert mpc(0,1)**3 == mpc(0,-1)
    assert mpc(0,1)**4 == 1
    assert mpc(0,1)**(10**9) == 1
    assert mpc(0,1)**(10**9+1) == mpc(0,1)
    assert mpc(0,1)**(10**9+2) == -1
    assert mpc(0,1)**(10**9+3) == mpc(0,-1)
    assert mpc(3,4)**10 == mpc(-9653287,1476984)
    assert mpc(mpq(1,2), mpq(-5,6))**3 == mpc(mpq(-11,12), mpq(-5,108))
    assert str(mpc(4,-3)) == '4 - 3*I'
    assert str(mpc(0,mpq(1,2))) == '1/2*I'
    assert mpc(0,1)**(-2) == -1
    assert abs(mpc(0,2)) == 2
    assert abs(mpc(mpf(3.0),mpf(4.0))) == 5.0

def test_extended_numbers():
    assert 1*oo == oo
    assert -3*oo == -oo
    assert oo*oo == oo
    assert oo+oo == oo
    assert mpq(1,3)*oo == oo
    assert -oo == (-1)*oo
    assert -(-oo) == oo
    assert oo - oo == nan
    assert -oo + -oo == -oo
    assert -oo - (-oo) == nan
    assert oo + nan == nan
    assert (mpc(0,1)*oo)*mpc(0,1) == -oo
    assert zoo * oo == zoo
    assert zoo * -oo == zoo
    assert zoo * nan == nan
    assert zoo + nan == nan
    assert 0*oo == nan
    assert 0*zoo == nan
    assert 0*nan == nan
    assert abs(oo) == abs(-oo) == abs(zoo) == oo
    assert abs(nan) == nan

def test_extended_cmp():
    assert -oo < oo
    assert -oo < 3
    assert -oo < mpq(-5,4)
    assert -oo <= 3
    #assert -oo <= -oo
    assert not (-oo > 3)
    assert not (-oo > -oo)
    assert not (-oo < -oo)
    assert oo > 3
    assert oo > -oo
    #assert oo <= oo
    assert not (oo < 3)
    assert not (oo < oo)
    assert not (oo > oo)
    assert not (nan < nan)
    assert not (nan > nan)
    assert max(2, 3, 1, nan, 2) == 3
    assert min(2, 3, 1, nan, 2) == 1
    assert max(2, -oo, oo, 3) == oo
    assert min(2, -oo, oo, 3) == -oo

def test_int_roots():
    assert int_root(1,1) == (1, True)
    assert int_root(0,1) == (0, True)
    assert int_root(0,3) == (0, True)
    assert int_root(10000, 1) == (10000, True)
    assert int_root(4,2) == (2, True)
    assert int_root(16,2) == (4, True)
    assert int_root(26,2) == (5, False)
    assert int_root(1234567**7, 7) == (1234567, True)
    assert int_root(1234567**7+1, 7) == (1234567, False)
    assert int_root(1234567**7-1, 7) == (1234566, False)
    b = 25**1000
    assert int_root(b, 1000) == (25, True)
    assert int_root(b+1, 1000) == (25, False)
    assert int_root(b-1, 1000) == (24, False)
    c = 10**400
    c2 = c**2
    assert int_root(c2, 2) == (c, True)
    assert int_root(c2+1, 2) == (c, False)
    assert int_root(c2-1, 2) == (c-1, False)
    assert int_root(2,10**10) == (1, False)

def test_powers():
    assert try_power(3, 2) == (9, [])
    assert try_power(3, -2) == (mpq(1, 9), [])
    assert try_power(0, -1) == (zoo, [])
    assert try_power(0, 0) == (1, [])
    assert try_power(mpq(1,2), 0) == (1, [])
    assert try_power(mpc(0, 1), 2) == (-1, [])
    assert try_power(mpc(-1,2), -2) == (mpc(mpq(-3,25), mpq(4,25)), [])
    assert try_power(2, mpq(1, 2)) == (1, [(2, mpq(1, 2))])
    assert try_power(4, mpq(1, 2)) == (2, [])
    assert try_power(4, mpq(3, 2)) == (8, [])
    assert try_power(4, mpq(-3, 2)) == (mpq(1,8), [])
    assert try_power(-1, mpq(1, 2)) == (mpc(0, 1), [])
    assert try_power(-729, mpq(1, 6)) == (3, [(-1, mpq(1, 6))])
    assert try_power(mpq(4,9), mpq(1,2)) == (mpq(2, 3), [])
    assert try_power(mpq(-4,9), mpq(1,2)) == (mpc(0, mpq(2, 3)), [])
    assert try_power(mpq(-4,9), mpq(-1,2)) == (mpc(0, mpq(-3, 2)), [])
    assert try_power(mpq(4,7), mpq(1,2)) == (2, [(7, mpq(-1, 2))])
    assert try_power(mpq(7,4), mpq(1,2)) == (mpq(1, 2), [(7, mpq(1, 2))])
    assert try_power(oo, 1) == (oo, [])
    assert try_power(oo, 3) == (oo, [])
    assert try_power(oo, 0) == (1, [])
    assert try_power(oo, -1) == (0, [])
    assert try_power(oo, -3) == (0, [])
    assert try_power(oo, mpq(-2,3)) == (0, [])
    assert try_power(-oo, 1) == (-oo, [])
    assert try_power(-oo, 2) == (oo, [])
    assert try_power(-oo, 3) == (-oo, [])
    assert try_power(-oo, mpq(5,2)) == (oo*mpc(0,1), [])
    #assert try_power(oo*mpc(0,1), mpq(5,2)) == (oo, [(mpc(0, 1), mpq(5, 2))])
    assert try_power(zoo, 2) == (zoo, [])
    assert try_power(-zoo, 3) == (zoo, [])
    assert try_power(zoo, -1) == (0, [])
